import argparse
import base64
import hashlib
import io
import json
import logging
import os
import subprocess
import sys
import tempfile
import time
import zipfile

import requests
from cryptography.hazmat.primitives import serialization as crypto_serialization

from keylime import api_version as keylime_api_version
from keylime import ca_util, config, crypto, keylime_logging, measured_boot, registrar_client, signing, web_util
from keylime.agentstates import AgentAttestState
from keylime.cmd import user_data_encrypt
from keylime.common import algorithms, retry, states, validators
from keylime.ima import file_signatures, ima
from keylime.requests_client import RequestsClient
from keylime.tpm import tpm2_objects
from keylime.tpm.tpm_abstract import TPM_Utilities
from keylime.tpm.tpm_main import tpm

# setup logging
logger = keylime_logging.init_logging("tenant")

# special exception that suppresses stack traces when it happens
class UserError(Exception):
    pass


class Tenant:
    """Simple command processor example."""

    config = None

    verifier_ip = None
    verifier_port = None

    agent_ip = None
    cv_cloudagent_ip = None
    agent_port = None

    registrar_ip = None
    registrar_port = None
    registrar_data = None

    api_version = None

    uuid_service_generate_locally = None
    agent_uuid = None

    K = None
    V = None
    U = None
    auth_tag = None

    tpm_policy = None
    metadata = {}
    allowlist = {}
    ima_policy_name = ""
    ima_sign_verification_keys = []
    revocation_key = ""
    accept_tpm_hash_algs = []
    accept_tpm_encryption_algs = []
    accept_tpm_signing_algs = []
    mb_refstate = None
    supported_version = None

    client_cert = None
    client_key = None
    client_key_password = None
    trusted_server_ca = False
    enable_agent_mtls = False
    verify_server_cert = False
    verify_custom = None

    request_timeout = None

    # Context with the trusted CA certificates from the configuration
    tls_context = None

    # Context with the agent's mTLS certificate
    agent_tls_context = None

    payload = None

    tpm_instance = tpm()

    def __init__(self):
        """Set up required values and TLS"""
        self.nonce = None
        self.agent_ip = None
        self.agent_port = None
        self.verifier_id = None
        self.verifier_ip = config.get("tenant", "verifier_ip")
        self.verifier_port = config.get("tenant", "verifier_port")
        self.registrar_ip = config.get("tenant", "registrar_ip")
        self.registrar_port = config.get("tenant", "registrar_port")
        self.api_version = keylime_api_version.current_version()
        self.enable_agent_mtls = config.getboolean("tenant", "enable_agent_mtls")
        self.request_timeout = config.getint("tenant", "request_timeout", fallback=60)

        logger.info("Setting up client TLS...")
        (cert, key, trusted_ca, key_password), verify_server_cert = web_util.get_tls_options(
            "tenant", is_client=True, logger=logger
        )

        if not self.enable_agent_mtls:
            logger.warning(
                "Warning: agent mTLS is currently disabled, keys will be sent in the clear! This should only be used for testing."
            )

        if not verify_server_cert:
            logger.warning(
                "Warning: server certificate verification is disabled as 'trusted_server_ca' option is set to 'all'. This should only be used for testing."
            )

        if not trusted_ca:
            logger.warning("No certificates provided in 'trusted_server_ca'")

        if cert and not os.path.isfile(cert):
            logger.warning("Could not find file %s provided in 'client_cert'", cert)

        if key and not os.path.isfile(key):
            logger.warning("Could not find file %s provided in 'client_key'", key)

        if all([(cert and os.path.isfile(cert)), (key and os.path.isfile(key))]):
            self.client_cert = cert
            self.client_key = key
            self.client_key_password = key_password
            self.verify_server_cert = verify_server_cert
            self.trusted_server_ca = trusted_ca

            self.tls_context = web_util.generate_tls_context(
                cert, key, trusted_ca, key_password, verify_server_cert, is_client=True, logger=logger
            )

            logger.info("TLS is enabled.")
        else:
            logger.warning("TLS is disabled.")

    @property
    def verifier_base_url(self):
        return f"{self.verifier_ip}:{self.verifier_port}"

    def process_allowlist(self, args):
        # Set up PCR values
        if "tpm_policy" in args and args["tpm_policy"] is not None:
            tpm_policy = args["tpm_policy"]
        else:
            # Use default empty TPM policy
            tpm_policy = "{}"
        self.tpm_policy = TPM_Utilities.readPolicy(tpm_policy)
        logger.info("TPM PCR Mask from policy is %s", self.tpm_policy["mask"])

        if len(args.get("ima_sign_verification_keys")) > 0:
            # Auto-enable IMA (or-bit mask)
            self.tpm_policy["mask"] = hex(int(self.tpm_policy["mask"], 0) | (1 << config.IMA_PCR))

            # Add all IMA file signing verification keys to a keyring
            tenant_keyring = file_signatures.ImaKeyring()
            for filename in args["ima_sign_verification_keys"]:
                pubkey, keyidv2 = file_signatures.get_pubkey_from_file(filename)
                if not pubkey:
                    raise UserError(f"File '{filename}' is not a file with a key")
                tenant_keyring.add_pubkey(pubkey, keyidv2)
            self.ima_sign_verification_keys = tenant_keyring.to_string()

        # Read command-line path string allowlist
        al_data = None

        if "allowlist" in args and args["allowlist"] is not None:

            self.enforce_pcrs(list(self.tpm_policy.keys()), [config.IMA_PCR], "IMA")

            # Auto-enable IMA (or-bit mask)
            self.tpm_policy["mask"] = hex(int(self.tpm_policy["mask"], 0) | (1 << config.IMA_PCR))

            try:
                al_data = ima.read_allowlist(
                    args["allowlist"], args["allowlist_checksum"], args["allowlist_sig"], args["allowlist_sig_key"]
                )
            except Exception as ima_e:
                raise UserError(str(ima_e)) from ima_e

        # Read command-line path string IMA exclude list
        excl_data = None
        if "ima_exclude" in args and args["ima_exclude"] is not None:
            if isinstance(args["ima_exclude"], str):
                excl_data = ima.read_excllist(args["ima_exclude"])
            elif isinstance(args["ima_exclude"], list):
                excl_data = args["ima_exclude"]
            else:
                raise UserError("Invalid exclude list provided")

        # Set up IMA
        if TPM_Utilities.check_mask(self.tpm_policy["mask"], config.IMA_PCR):
            # Process allowlists
            al_data["excllist"] = excl_data
            self.allowlist = al_data

        # Store allowlist name
        if "allowlist_name" in args and args["allowlist_name"] is not None:
            self.ima_policy_name = args["allowlist_name"]
            # Auto-enable IMA (or-bit mask)
            self.tpm_policy["mask"] = hex(int(self.tpm_policy["mask"], 0) | (1 << config.IMA_PCR))

        # Read command-line path string TPM event log (measured boot) reference state
        mb_refstate_data = None
        if "mb_refstate" in args and args["mb_refstate"] is not None:

            self.enforce_pcrs(list(self.tpm_policy.keys()), config.MEASUREDBOOT_PCRS, "measured boot")

            # Auto-enable TPM event log mesured boot (or-bit mask)
            for _pcr in config.MEASUREDBOOT_PCRS:
                self.tpm_policy["mask"] = hex(int(self.tpm_policy["mask"], 0) | (1 << _pcr))

            logger.info(
                "TPM PCR Mask automatically modified is %s to include IMA/Event log PCRs", self.tpm_policy["mask"]
            )

            if isinstance(args["mb_refstate"], str):
                if args["mb_refstate"] == "default":
                    args["mb_refstate"] = config.get("tenant", "mb_refstate")
                mb_refstate_data = measured_boot.read_mb_refstate(args["mb_refstate"])
            else:
                raise UserError("Invalid measured boot reference state (intended state) provided")

        # Set up measured boot (TPM event log) reference state
        if TPM_Utilities.check_mask(self.tpm_policy["mask"], config.MEASUREDBOOT_PCRS[2]):
            # Process measured boot reference state
            self.mb_refstate = mb_refstate_data

    def init_add(self, args):
        """Set up required values. Command line options can overwrite these config values

        Arguments:
            args {[string]} -- agent_ip|agent_port|cv_agent_ip
        """
        if "agent_ip" in args:
            self.agent_ip = args["agent_ip"]

        if "agent_port" in args and args["agent_port"] is not None:
            self.agent_port = args["agent_port"]

        self.registrar_data = registrar_client.getData(
            self.registrar_ip, self.registrar_port, self.agent_uuid, self.tls_context
        )

        if self.registrar_data is None:
            raise UserError(f"Agent ${self.agent_uuid} data not found in the Registrar.")

        # try to get the port or ip from the registrar if it is missing
        if (self.agent_ip is None or self.agent_port is None) and self.registrar_data is not None:
            if self.agent_ip is None:
                if self.registrar_data["ip"] is not None:
                    self.agent_ip = self.registrar_data["ip"]
                else:
                    raise UserError("No Ip was specified or found in the Registrar")

            if self.agent_port is None and self.registrar_data["port"] is not None:
                self.agent_port = self.registrar_data["port"]

        # Check if a contact ip and port for the agent was found
        if self.agent_ip is None:
            raise UserError("The contact ip address for the agent was not specified.")

        if self.agent_port is None:
            raise UserError("The contact port for the agent was not specified.")

        # Auto-detection for API version
        self.supported_version = args["supported_version"]
        if self.supported_version is None:
            # Default to 1.0 if the agent did not send a mTLS certificate
            if self.registrar_data.get("mtls_cert", None) is None:
                self.supported_version = "1.0"
            else:
                # Try to connect to the agent to get supported version
                if self.registrar_data["mtls_cert"] == "disabled":
                    self.enable_agent_mtls = False
                    logger.warning(
                        "Warning: mTLS for agents is disabled: the identity of each node will be based on the properties of the TPM only. "
                        "Unless you have strict control of your network, it is strongly advised that remote code execution should be disabled, "
                        'by setting "payload_script=" and "extract_payload_zip=False" under "[agent]" in agent configuration file.'
                    )
                    tls_context = None
                else:
                    # Store the agent self-signed certificate as a string
                    self.verify_custom = self.registrar_data["mtls_cert"]

                    if not self.agent_tls_context:
                        self.agent_tls_context = web_util.generate_tls_context(
                            self.client_cert,
                            self.client_key,
                            self.trusted_server_ca,
                            self.client_key_password,
                            self.verify_server_cert,
                            is_client=True,
                            ca_cert_string=self.verify_custom,
                            logger=logger,
                        )
                    tls_context = self.agent_tls_context

                with RequestsClient(
                    f"{self.agent_ip}:{self.agent_port}",
                    tls_enabled=self.enable_agent_mtls,
                    tls_context=tls_context,
                ) as get_version:
                    res = get_version.get("/version")
                    if res and res.status_code == 200:
                        try:
                            data = res.json()
                            api_version = data["results"]["supported_version"]
                            if keylime_api_version.validate_version(api_version):
                                self.supported_version = api_version
                            else:
                                logger.warning("API version provided by the agent is not valid")
                        except (TypeError, KeyError):
                            pass

        if self.supported_version is None:
            api_version = keylime_api_version.current_version()
            logger.warning("Could not detect supported API version. Defaulting to %s", api_version)
            self.supported_version = api_version

        # Now set the cv_agent_ip
        if "cv_agent_ip" in args and args["cv_agent_ip"] is not None:
            self.cv_cloudagent_ip = args["cv_agent_ip"]
        else:
            self.cv_cloudagent_ip = self.agent_ip

        # Make sure all keys exist in dictionary
        if "file" not in args:
            args["file"] = None
        if "keyfile" not in args:
            args["keyfile"] = None
        if "payload" not in args:
            args["payload"] = None
        if "ca_dir" not in args:
            args["ca_dir"] = None
        if "incl_dir" not in args:
            args["incl_dir"] = None
        if "ca_dir_pw" not in args:
            args["ca_dir_pw"] = None

        # Set up accepted algorithms
        self.accept_tpm_hash_algs = config.getlist("tenant", "accept_tpm_hash_algs")
        self.accept_tpm_encryption_algs = config.getlist("tenant", "accept_tpm_encryption_algs")
        self.accept_tpm_signing_algs = config.getlist("tenant", "accept_tpm_signing_algs")

        self.process_allowlist(args)

        # if none
        if args["file"] is None and args["keyfile"] is None and args["ca_dir"] is None:
            raise UserError(
                "You must specify one of -k, -f, or --cert to specify the key/contents to be securely delivered to the agent"
            )

        if args["keyfile"] is not None:
            if args["file"] is not None or args["ca_dir"] is not None:
                raise UserError(
                    "You must specify one of -k, -f, or --cert to specify the key/contents to be securely delivered to the agent"
                )

            # read the keys in
            if isinstance(args["keyfile"], dict) and "data" in args["keyfile"]:
                if isinstance(args["keyfile"]["data"], list) and len(args["keyfile"]["data"]) == 1:
                    keyfile = args["keyfile"]["data"][0]
                    if keyfile is None:
                        raise UserError("Invalid key file contents")
                    f = io.StringIO(keyfile)
                else:
                    raise UserError("Invalid key file provided")
            else:
                f = open(args["keyfile"], encoding="utf-8")  # pylint: disable=consider-using-with
            self.K = base64.b64decode(f.readline())
            self.U = base64.b64decode(f.readline())
            self.V = base64.b64decode(f.readline())
            f.close()

            # read the payload in (opt.)
            if isinstance(args["payload"], dict) and "data" in args["payload"]:
                if isinstance(args["payload"]["data"], list) and len(args["payload"]["data"]) > 0:
                    self.payload = args["payload"]["data"][0]
            else:
                if args["payload"] is not None:
                    with open(args["payload"], "rb") as f:
                        self.payload = f.read()

        if args["file"] is not None:
            if args["keyfile"] is not None or args["ca_dir"] is not None:
                raise UserError(
                    "You must specify one of -k, -f, or --cert to specify the key/contents to be securely delivered to the agent"
                )

            if isinstance(args["file"], dict) and "data" in args["file"]:
                if isinstance(args["file"]["data"], list) and len(args["file"]["data"]) > 0:
                    contents = args["file"]["data"][0]
                    if contents is None:
                        raise UserError("Invalid file payload contents")
                else:
                    raise UserError("Invalid file payload provided")
            else:
                with open(args["file"], encoding="utf-8") as f:
                    contents = f.read()
            ret = user_data_encrypt.encrypt(contents)
            self.K = ret["k"]
            self.U = ret["u"]
            self.V = ret["v"]
            self.payload = ret["ciphertext"]

        if args["ca_dir"] is None and args["incl_dir"] is not None:
            raise UserError("--include option is only valid when used with --cert")
        if args["ca_dir"] is not None:
            if args["file"] is not None or args["keyfile"] is not None:
                raise UserError(
                    "You must specify one of -k, -f, or --cert to specify the key/contents to be securely delivered to the agent"
                )
            if args["ca_dir"] == "default":
                args["ca_dir"] = config.CA_WORK_DIR

            if "ca_dir_pw" in args and args["ca_dir_pw"] is not None:
                ca_util.setpassword(args["ca_dir_pw"])

            if not os.path.exists(args["ca_dir"]) or not os.path.exists(os.path.join(args["ca_dir"], "cacert.crt")):
                logger.warning("CA directory does not exist. Creating...")
                ca_util.cmd_init(args["ca_dir"])
            if not os.path.exists(os.path.join(args["ca_dir"], f"{self.agent_uuid}-private.pem")):
                ca_util.cmd_mkcert(args["ca_dir"], self.agent_uuid)

            cert_pkg, serial, subject = ca_util.cmd_certpkg(args["ca_dir"], self.agent_uuid)

            # support revocation
            if not os.path.exists(os.path.join(args["ca_dir"], "RevocationNotifier-private.pem")):
                ca_util.cmd_mkcert(args["ca_dir"], "RevocationNotifier")
            rev_package, _, _ = ca_util.cmd_certpkg(args["ca_dir"], "RevocationNotifier")

            # extract public and private keys from package
            sf = io.BytesIO(rev_package)
            with zipfile.ZipFile(sf) as zf:
                privkey = zf.read("RevocationNotifier-private.pem")
                cert = zf.read("RevocationNotifier-cert.crt")

            # put the cert of the revoker into the cert package
            sf = io.BytesIO(cert_pkg)
            with zipfile.ZipFile(sf, "a", compression=zipfile.ZIP_STORED) as zf:
                zf.writestr("RevocationNotifier-cert.crt", cert)

                # add additional files to zip
                if args["incl_dir"] is not None:
                    if isinstance(args["incl_dir"], dict) and "data" in args["incl_dir"] and "name" in args["incl_dir"]:
                        if isinstance(args["incl_dir"]["data"], list) and isinstance(args["incl_dir"]["name"], list):
                            if len(args["incl_dir"]["data"]) != len(args["incl_dir"]["name"]):
                                raise UserError("Invalid incl_dir provided")
                            for i in range(len(args["incl_dir"]["data"])):
                                zf.writestr(os.path.basename(args["incl_dir"]["name"][i]), args["incl_dir"]["data"][i])
                    else:
                        if os.path.exists(args["incl_dir"]):
                            files = next(os.walk(args["incl_dir"]))[2]
                            for filename in files:
                                with open(os.path.join(args["incl_dir"], filename), "rb") as f:
                                    zf.writestr(os.path.basename(f.name), f.read())
                        else:
                            logger.warning(
                                "Specified include directory %s does not exist. Skipping...", args["incl_dir"]
                            )

            cert_pkg = sf.getvalue()

            # put the private key into the data to be send to the CV
            self.revocation_key = privkey.decode("utf-8")

            # encrypt up the cert package
            ret = user_data_encrypt.encrypt(cert_pkg)
            self.K = ret["k"]
            self.U = ret["u"]
            self.V = ret["v"]
            self.metadata = {"cert_serial": serial, "subject": subject}
            self.payload = ret["ciphertext"]

        if self.payload is not None:
            max_payload_size = config.getint("tenant", "max_payload_size")
            if len(self.payload) > max_payload_size:
                raise UserError(f"Payload size {len(self.payload)} exceeds max size {max_payload_size}")

    def enforce_pcrs(self, policy_pcrs, protected_pcrs, pcr_use):
        policy_pcrs = list(self.tpm_policy.keys())
        policy_pcrs.remove("mask")

        for _pcr in policy_pcrs:
            if int(_pcr) in protected_pcrs:
                logger.error(
                    'WARNING: PCR %s is specified in "tpm_policy", but will in fact be used by %s. Please remove it from policy',
                    _pcr,
                    pcr_use,
                )
                sys.exit(1)

    def preloop(self):
        """encrypt the agent UUID as a check for delivering the correct key"""
        self.auth_tag = crypto.do_hmac(self.K, self.agent_uuid)
        # be very careful printing K, U, or V as they leak in logs stored on unprotected disks
        if config.INSECURE_DEBUG:
            logger.debug("K: %s", base64.b64encode(self.K))
            logger.debug("V: %s", base64.b64encode(self.V))
            logger.debug("U: %s", base64.b64encode(self.U))
            logger.debug("Auth Tag: %s", self.auth_tag)

    def check_ek(self, ekcert):
        """Check the Entity Key

        Arguments:
            ekcert {str} -- The endorsement key, either None, "emulator", or base64 encoded der cert

        Returns:
            [type] -- [description]
        """
        if config.getboolean("tenant", "require_ek_cert"):
            if ekcert == "emulator" and config.DISABLE_EK_CERT_CHECK_EMULATOR:
                logger.info("Not checking ekcert of TPM emulator")
            elif ekcert is None:
                logger.warning("No EK cert provided, require_ek_cert option in config set to True")
                return False
            elif not self.tpm_instance.verify_ek(base64.b64decode(ekcert)):
                logger.warning("Invalid EK certificate")
                return False

        return True

    def validate_tpm_quote(self, public_key, quote, hash_alg):
        """Validate TPM Quote received from the Agent

        Arguments:
            public_key {[type]} -- [description]
            quote {[type]} -- [description]
            hash_alg {bool} -- [description]

        Raises:
            UserError: [description]

        Returns:
            [type] -- [description]
        """
        if self.registrar_data is None:
            logger.warning("AIK not found in registrar, quote not validated")
            return False

        failure = self.tpm_instance.check_quote(
            AgentAttestState(self.agent_uuid),
            self.nonce,
            public_key,
            quote,
            self.registrar_data["aik_tpm"],
            hash_alg=hash_alg,
            compressed=(self.supported_version == "1.0"),
        )
        if failure:
            if self.registrar_data["regcount"] > 1:
                logger.error(
                    "WARNING: This UUID had more than one ek-ekcert registered to it! This might indicate that your system is misconfigured or a malicious host is present. Run 'regdelete' for this agent and restart"
                )
                sys.exit()
            return False

        if self.registrar_data["regcount"] > 1:
            logger.warning(
                "WARNING: This UUID had more than one ek-ekcert registered to it! This might indicate that your system is misconfigured. Run 'regdelete' for this agent and restart"
            )

        if not config.getboolean("tenant", "require_ek_cert") and config.get("tenant", "ek_check_script") == "":
            logger.warning(
                "DANGER: EK cert checking is disabled and no additional checks on EKs have been specified with ek_check_script option. Keylime is not secure!!"
            )
            return True

        # check EK cert and make sure it matches EK
        if not self.check_ek(self.registrar_data["ekcert"]):
            return False
        # if agent is virtual, check phyisical EK cert and make sure it matches phyiscal EK
        if "provider_keys" in self.registrar_data:
            if not self.check_ek(self.registrar_data["provider_keys"]["ekcert"]):
                return False

        # check all EKs with optional script:
        script = config.get("tenant", "ek_check_script")
        if not script:
            return True

        if script[0] != "/":
            script = os.path.join(config.WORK_DIR, script)

        logger.info("Checking EK with script %s", script)
        # now we need to exec the script with the ek and ek cert in vars
        env = os.environ.copy()
        env["AGENT_UUID"] = self.agent_uuid
        env["EK"] = tpm2_objects.pubkey_from_tpm2b_public(
            base64.b64decode(self.registrar_data["ek_tpm"]),
        ).public_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        env["EK_TPM"] = self.registrar_data["ek_tpm"]
        if self.registrar_data["ekcert"] is not None:
            env["EK_CERT"] = self.registrar_data["ekcert"]
        else:
            env["EK_CERT"] = ""

        env["PROVKEYS"] = json.dumps(self.registrar_data.get("provider_keys", {}))
        with subprocess.Popen(
            script, env=env, shell=True, cwd=config.WORK_DIR, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ) as proc:
            retval = proc.wait()
            if retval != 0:
                raise UserError("External check script failed to validate EK")
            logger.debug("External check script successfully to validated EK")
            while True:
                line = proc.stdout.readline().decode()
                if line == "":
                    break
                logger.debug("ek_check output: %s", line.strip())
        return True

    def do_cv(self):
        """Initiate v, agent_id and ip and initiate the cloudinit sequence"""
        b64_v = base64.b64encode(self.V).decode("utf-8")
        logger.debug("b64_v: %s", b64_v)
        data = {
            "v": b64_v,
            "cloudagent_ip": self.cv_cloudagent_ip,
            "cloudagent_port": self.agent_port,
            "tpm_policy": json.dumps(self.tpm_policy),
            "ima_policy_bundle": json.dumps(self.allowlist),
            "ima_policy_name": self.ima_policy_name,
            "mb_refstate": json.dumps(self.mb_refstate),
            "ima_sign_verification_keys": json.dumps(self.ima_sign_verification_keys),
            "metadata": json.dumps(self.metadata),
            "revocation_key": self.revocation_key,
            "accept_tpm_hash_algs": self.accept_tpm_hash_algs,
            "accept_tpm_encryption_algs": self.accept_tpm_encryption_algs,
            "accept_tpm_signing_algs": self.accept_tpm_signing_algs,
            "ak_tpm": self.registrar_data["aik_tpm"],
            "mtls_cert": self.registrar_data.get("mtls_cert", None),
            "supported_version": self.supported_version,
        }
        json_message = json.dumps(data)
        do_cv = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = do_cv.post(
            (f"/v{self.api_version}/agents/{self.agent_uuid}"), data=json_message, timeout=self.request_timeout
        )

        if response.status_code == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            sys.exit()
        elif response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            sys.exit()

        if response.status_code == 409:
            # this is a conflict, need to update or delete it
            Tenant._print_json_response(response)
            sys.exit()
        elif response.status_code != 200:
            keylime_logging.log_http_response(logger, logging.ERROR, response.json())
            logger.error(
                "POST command response: %s Unexpected response from Cloud Verifier: %s",
                response.status_code,
                response.text,
            )
            sys.exit()

    def do_cvstatus(self):
        """Perform operational state look up for agent on the verifier"""

        do_cvstatus = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)

        response = do_cvstatus.get((f"/v{self.api_version}/agents/{self.agent_uuid}"), timeout=self.request_timeout)

        if response.status_code == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            return response.json()
        if response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            return response.json()
        if response.status_code == 404:
            logger.info(
                "Verifier at %s with Port %s does not have agent %s.",
                self.verifier_ip,
                self.verifier_port,
                self.agent_uuid,
            )
            return response.json()
        if response.status_code == 200:
            response = response.json()

            res = response.pop("results")
            response["results"] = {self.agent_uuid: res}

            operational_state = states.state_to_str(response["results"][self.agent_uuid]["operational_state"])
            response["results"][self.agent_uuid]["operational_state"] = operational_state

            logger.info("Agent Info:\n%s", json.dumps(response["results"]))

            return response

        logger.info(
            "Status command response: %s. Unexpected response from Cloud Verifier %s on port %s. %s",
            response.status_code,
            self.verifier_ip,
            self.verifier_port,
            str(response),
        )
        return response

    def do_cvlist(self):
        """List all agent statuses in cloudverifier"""

        do_cvstatus = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)

        verifier_id = ""
        if self.verifier_id is not None:
            verifier_id = self.verifier_id

        response = do_cvstatus.get(f"/v{self.api_version}/agents/?verifier={verifier_id}", timeout=self.request_timeout)

        if response.status_code == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            return response.json()
        if response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            return response.json()
        if response.status_code == 404:
            logger.info(
                "Verifier at %s with Port %s does not have agent %s.",
                self.verifier_ip,
                self.verifier_port,
                self.agent_uuid,
            )
            return response.json()
        if response.status_code == 200:
            response = response.json()

            logger.info('From verifier %s port %s retrieved: "%s"', self.verifier_ip, self.verifier_port, response)

            return response

        logger.info(
            "Status command response: %s. Unexpected response from Cloud Verifier %s on port %s. %s",
            response.status_code,
            self.verifier_ip,
            self.verifier_port,
            str(response),
        )
        return response

    def do_cvbulkinfo(self):
        """Perform operational state look up for agent"""

        do_cvstatus = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)

        verifier_id = ""
        if self.verifier_id is not None:
            verifier_id = self.verifier_id

        response = do_cvstatus.get(
            f"/v{self.api_version}/agents/?bulk={True}&verifier={verifier_id}", timeout=self.request_timeout
        )

        if response.status_code == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            return response.json()
        if response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            return response.json()
        if response.status_code == 404:
            logger.info(
                "Verifier at %s with Port %s does not have agent %s.",
                self.verifier_ip,
                self.verifier_port,
                self.agent_uuid,
            )
            return response.json()
        if response.status_code == 200:
            response = response.json()

            for agent in response["results"].keys():
                response["results"][agent]["operational_state"] = states.state_to_str(
                    response["results"][agent]["operational_state"]
                )
            logger.info("Bulk Agent Info:\n%s", json.dumps(response["results"]))

            return response

        logger.info(
            "Status command response: %s. Unexpected response from Cloud Verifier %s on port %s. %s",
            response.status_code,
            self.verifier_ip,
            self.verifier_port,
            str(response),
        )
        return response

    def do_cvdelete(self, verifier_check=True):
        """Delete agent from Verifier."""
        if verifier_check:
            cvresponse = self.do_cvstatus()

            if not isinstance(cvresponse, dict):
                return cvresponse

            if cvresponse["code"] != 200:
                logger.error("Could not get status of agent %s from verifier %s.", self.agent_uuid, self.verifier_ip)
                return cvresponse

            self.verifier_ip = cvresponse["results"][self.agent_uuid]["verifier_ip"]
            self.verifier_port = cvresponse["results"][self.agent_uuid]["verifier_port"]

        do_cvdelete = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = do_cvdelete.delete(f"/v{self.api_version}/agents/{self.agent_uuid}", timeout=self.request_timeout)

        response = response.json()

        if response["code"] == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            return response
        if response["code"] == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            return response
        if response["code"] == 202:
            deleted = False
            for _ in range(12):
                get_cvdelete = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
                response = get_cvdelete.get(
                    f"/v{self.api_version}/agents/{self.agent_uuid}", timeout=self.request_timeout
                )

                if response.status_code == 404:
                    deleted = True
                    break
                time.sleep(0.4)
            if deleted:
                logger.info("Agent %s deleted from the CV", self.agent_uuid)
                return response.json()
            logger.error("Timed out waiting for delete of agent %s to complete at CV", self.agent_uuid)
            return response.json()
        if response["code"] == 200:
            logger.info("Agent %s deleted from the CV", self.agent_uuid)
            return response

        keylime_logging.log_http_response(logger, logging.ERROR, response)
        return response

    def do_regstatus(self):
        agent_info = registrar_client.getData(self.registrar_ip, self.registrar_port, self.agent_uuid, self.tls_context)

        if not agent_info:
            logger.info(
                "Agent %s does not exist on the registrar. Please register the agent with the registrar.",
                self.agent_uuid,
            )
            response = {
                "code": 404,
                "status": f"Agent {self.agent_uuid} does not exist on "
                f"registrar {self.registrar_ip} port {self.registrar_port}.",
                "results": {},
            }
            logger.info(json.dumps(response))
            return response

        response = {
            "code": 200,
            "status": f"Agent {self.agent_uuid} exists on "
            f"registrar {self.registrar_ip} port {self.registrar_port}.",
            "results": {},
        }
        response["results"][self.agent_uuid] = agent_info
        response["results"][self.agent_uuid]["operational_state"] = states.state_to_str(states.REGISTERED)

        logger.info(json.dumps(response))

        return response

    def do_reglist(self):
        """List agents from Registrar"""
        response = registrar_client.doRegistrarList(
            self.registrar_ip, self.registrar_port, tls_context=self.tls_context
        )

        logger.info(
            "From registrar %s port %s retrieved %s", self.registrar_ip, self.registrar_port, json.dumps(response)
        )
        return response

    def do_regdelete(self):
        """Delete agent from Registrar"""
        response = registrar_client.doRegistrarDelete(
            self.registrar_ip, self.registrar_port, self.agent_uuid, tls_context=self.tls_context
        )

        return response

    def do_status(self):
        """Perform operational state look up for agent"""

        regresponse = self.do_regstatus()

        if regresponse["code"] == 404:
            return regresponse

        cvresponse = self.do_cvstatus()

        if not isinstance(cvresponse, dict):
            logger.error(
                "Unexpected response from Cloud Verifier %s on port %s. response %s",
                self.verifier_ip,
                self.verifier_port,
                str(cvresponse),
            )
            return cvresponse

        if regresponse["code"] == 200 and cvresponse["code"] == 200:
            return cvresponse
        if regresponse["code"] == 200 and cvresponse["code"] != 200:
            return regresponse

        logger.error(
            "Unknown inconsistent state between registrar %s on "
            "port %s and verifier %s on port %s occured. Got "
            "registrar response %s verifier response %s",
            self.verifier_ip,
            self.verifier_port,
            self.registrar_ip,
            self.registrar_port,
            str(regresponse),
            str(cvresponse),
        )

        return {"registrar": regresponse, "verifier": cvresponse}

    def do_cvreactivate(self, verifier_check=True):
        """Reactive Agent."""
        if verifier_check:
            agent_json = self.do_cvstatus()
            self.verifier_ip = agent_json["results"][self.agent_uuid]["verifier_ip"]
            self.verifier_port = agent_json["results"][self.agent_uuid]["verifier_port"]

        do_cvreactivate = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = do_cvreactivate.put(
            f"/v{self.api_version}/agents/{self.agent_uuid}/reactivate",
            data=b"",
            timeout=self.request_timeout,
        )

        if response.status_code == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            return response.json()
        if response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            return response.json()
        if response.status_code == 200:
            logger.info("Agent %s re-activated", self.agent_uuid)
            return response.json()

        response_body = response.json()
        keylime_logging.log_http_response(logger, logging.ERROR, response_body)
        logger.error("Update command response: %s Unexpected response from Cloud Verifier.", response.status_code)
        return response.json()

    def do_cvstop(self):
        """Stop declared active agent"""
        params = f"/v{self.api_version}/agents/{self.agent_uuid}/stop"
        do_cvstop = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = do_cvstop.put(params, data=b"", timeout=self.request_timeout)
        if response.status_code == 503:
            logger.error(
                "Cannot connect to Verifier at %s with Port %s. Connection refused.",
                self.verifier_ip,
                self.verifier_port,
            )
            sys.exit()
        elif response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            sys.exit()

        response_body = response.json()
        if response.status_code != 200:
            keylime_logging.log_http_response(logger, logging.ERROR, response_body)
        else:
            logger.info("Agent %s stopped", self.agent_uuid)

    def do_quote(self):
        """Perform TPM quote by GET towards Agent

        Raises:
            UserError: Connection handler
        """
        self.nonce = TPM_Utilities.random_password(20)

        numtries = 0
        response = None
        # Note: We need a specific retry handler (perhaps in common), no point having localised unless we have too.
        while True:
            try:
                params = f"/v{self.supported_version}/quotes/identity?nonce=%s" % (self.nonce)
                cloudagent_base_url = f"{self.agent_ip}:{self.agent_port}"

                if self.enable_agent_mtls and self.registrar_data and self.registrar_data["mtls_cert"]:
                    with RequestsClient(
                        cloudagent_base_url,
                        self.enable_agent_mtls,
                        tls_context=self.agent_tls_context,
                    ) as do_quote:
                        response = do_quote.get(params, timeout=self.request_timeout)
                else:
                    logger.warning("Connecting to agent without using mTLS!")
                    do_quote = RequestsClient(cloudagent_base_url, tls_enabled=False)
                    response = do_quote.get(params, timeout=self.request_timeout)

                print(response)
                response_body = response.json()

            except Exception as e:
                if response.status_code in (503, 504):
                    numtries += 1
                    maxr = config.getint("tenant", "max_retries")
                    if numtries >= maxr:
                        logger.error(
                            "Tenant cannot establish connection to agent on %s with port %s",
                            self.agent_ip,
                            self.agent_port,
                        )
                        sys.exit()
                    interval = config.getfloat("tenant", "retry_interval")
                    exponential_backoff = config.getboolean("tenant", "exponential_backoff")
                    next_retry = retry.retry_time(exponential_backoff, interval, numtries, logger)
                    logger.info(
                        "Tenant connection to agent at %s refused %s/%s times, trying again in %s seconds...",
                        self.agent_ip,
                        numtries,
                        maxr,
                        next_retry,
                    )
                    time.sleep(next_retry)
                    continue

                raise e
            break

        if response is not None and response.status_code != 200:
            raise UserError(f"Status command response: {response.status_code} Unexpected response from Cloud Agent.")

        if "results" not in response_body:
            raise UserError(f"Error: unexpected http response body from Cloud Agent: {str(response.status)}")

        quote = response_body["results"]["quote"]
        logger.debug("Agent_quote received quote: %s", quote)

        public_key = response_body["results"]["pubkey"]
        logger.debug("Agent_quote received public key: %s", public_key)

        # Ensure hash_alg is in accept_tpm_hash_algs list
        hash_alg = response_body["results"]["hash_alg"]
        logger.debug("Agent_quote received hash algorithm: %s", hash_alg)
        if not algorithms.is_accepted(
            hash_alg, config.getlist("tenant", "accept_tpm_hash_algs")
        ) or not algorithms.Hash.is_recognized(hash_alg):
            raise UserError(f"TPM Quote is using an unaccepted hash algorithm: {hash_alg}")

        # Ensure enc_alg is in accept_tpm_encryption_algs list
        enc_alg = response_body["results"]["enc_alg"]
        logger.debug("Agent_quote received encryption algorithm: %s", enc_alg)
        if not algorithms.is_accepted(enc_alg, config.getlist("tenant", "accept_tpm_encryption_algs")):
            raise UserError(f"TPM Quote is using an unaccepted encryption algorithm: {enc_alg}")

        # Ensure sign_alg is in accept_tpm_encryption_algs list
        sign_alg = response_body["results"]["sign_alg"]
        logger.debug("Agent_quote received signing algorithm: %s", sign_alg)
        if not algorithms.is_accepted(sign_alg, config.getlist("tenant", "accept_tpm_signing_algs")):
            raise UserError(f"TPM Quote is using an unaccepted signing algorithm: {sign_alg}")

        if not self.validate_tpm_quote(public_key, quote, algorithms.Hash(hash_alg)):
            raise UserError(f"TPM Quote from cloud agent is invalid for nonce: {self.nonce}")

        logger.info("Quote from %s validated", self.agent_ip)

        # encrypt U with the public key
        encrypted_U = crypto.rsa_encrypt(crypto.rsa_import_pubkey(public_key), self.U)

        b64_encrypted_u = base64.b64encode(encrypted_U)
        logger.debug("b64_encrypted_u: %s", b64_encrypted_u.decode("utf-8"))
        data = {"encrypted_key": b64_encrypted_u.decode("utf-8"), "auth_tag": self.auth_tag}

        if self.payload is not None:
            data["payload"] = self.payload.decode("utf-8")

        # post encrypted U back to CloudAgent
        params = f"/v{self.supported_version}/keys/ukey"
        cloudagent_base_url = f"{self.agent_ip}:{self.agent_port}"

        if self.enable_agent_mtls and self.registrar_data and self.registrar_data["mtls_cert"]:
            with RequestsClient(
                cloudagent_base_url,
                self.enable_agent_mtls,
                tls_context=self.agent_tls_context,
            ) as post_ukey:
                response = post_ukey.post(params, json=data, timeout=self.request_timeout)
        else:
            logger.warning("Connecting to agent without using mTLS!")
            post_ukey = RequestsClient(cloudagent_base_url, tls_enabled=False)
            response = post_ukey.post(params, json=data, timeout=self.request_timeout)

        if response.status_code == 503:
            logger.error(
                "Cannot connect to Agent at %s with Port %s. Connection refused.", self.agent_ip, self.agent_port
            )
            sys.exit()
        elif response.status_code == 504:
            logger.error("Verifier at %s with Port %s timed out.", self.verifier_ip, self.verifier_port)
            sys.exit()

        if response.status_code != 200:
            keylime_logging.log_http_response(logger, logging.ERROR, response_body)
            raise UserError(
                f"Posting of Encrypted U to the Cloud Agent failed with response code {response.status_code} ({response.text})"
            )

    def do_verify(self):
        """Perform verify using a random generated challenge"""
        challenge = TPM_Utilities.random_password(20)
        numtries = 0

        while True:
            response = None
            try:
                cloudagent_base_url = f"{self.agent_ip}:{self.agent_port}"

                if self.registrar_data and self.registrar_data["mtls_cert"]:
                    with RequestsClient(
                        cloudagent_base_url,
                        self.enable_agent_mtls,
                        tls_context=self.agent_tls_context,
                    ) as do_verify:
                        response = do_verify.get(
                            f"/v{self.supported_version}/keys/verify?challenge={challenge}",
                            timeout=self.request_timeout,
                        )
                else:
                    logger.warning("Connecting to agent without using mTLS!")
                    do_verify = RequestsClient(cloudagent_base_url, tls_enabled=False)
                    response = do_verify.get(
                        f"/v{self.supported_version}/keys/verify?challenge={challenge}", timeout=self.request_timeout
                    )

                response_body = response.json()
            except Exception as e:
                if response is not None and response.status_code in (503, 504):
                    numtries += 1
                    maxr = config.getint("tenant", "max_retries")
                    if numtries >= maxr:
                        logger.error(
                            "Cannot establish connection to agent on %s with port %s", self.agent_ip, self.agent_port
                        )
                        self.do_cvstop()
                        sys.exit()
                    interval = config.getfloat("tenant", "retry_interval")
                    exponential_backoff = config.getboolean("tenant", "exponential_backoff")
                    next_retry = retry.retry_time(exponential_backoff, interval, numtries, logger)
                    logger.info(
                        "Verifier connection to agent at %s refused %s/%s times, trying again in %s seconds...",
                        self.agent_ip,
                        numtries,
                        maxr,
                        next_retry,
                    )
                    time.sleep(next_retry)
                    continue
                self.do_cvstop()
                raise e
            if response.status_code == 200:
                if "results" not in response_body or "hmac" not in response_body["results"]:
                    logger.critical("Error: unexpected http response body from Cloud Agent: %s", response.status_code)
                    self.do_cvstop()
                    break
                mac = response_body["results"]["hmac"]

                ex_mac = crypto.do_hmac(self.K, challenge)

                if mac == ex_mac:
                    logger.info("Key derivation successful")
                else:
                    logger.error("Key derivation failed")
                    self.do_cvstop()
            else:
                keylime_logging.log_http_response(logger, logging.ERROR, response_body)
                numtries += 1
                maxr = config.getint("tenant", "max_retries")
                if numtries >= maxr:
                    logger.error("Agent on %s with port %s failed key derivation", self.agent_ip, self.agent_port)
                    self.do_cvstop()
                    sys.exit()
                interval = config.getfloat("tenant", "retry_interval")
                exponential_backoff = config.getboolean("tenant", "exponential_backoff")
                next_retry = retry.retry_time(exponential_backoff, interval, numtries, logger)
                logger.info(
                    "Key derivation not yet complete (retry %s/%s), trying again in %s seconds... (Ctrl-C to stop)",
                    numtries,
                    maxr,
                    next_retry,
                )
                time.sleep(next_retry)
                continue
            break

    def do_add_allowlist(self, args):
        if "allowlist_name" not in args or not args["allowlist_name"]:
            raise UserError("allowlist_name is required to add an allowlist")

        self.process_allowlist(args)
        data = {"tpm_policy": json.dumps(self.tpm_policy), "ima_policy_bundle": json.dumps(self.allowlist)}
        body = json.dumps(data)
        cv_client = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = cv_client.post(
            f"/v{self.api_version}/allowlists/{self.ima_policy_name}", data=body, timeout=self.request_timeout
        )
        Tenant._print_json_response(response)

    def do_delete_allowlist(self, name):
        cv_client = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = cv_client.delete(f"/v{self.api_version}/allowlists/{name}", timeout=self.request_timeout)
        Tenant._print_json_response(response)

    def do_show_allowlist(self, name):  # pylint: disable=unused-argument
        cv_client = RequestsClient(self.verifier_base_url, True, tls_context=self.tls_context)
        response = cv_client.get(f"/v{self.api_version}/allowlists/{name}", timeout=self.request_timeout)
        print(f"Show allowlist command response: {response.status_code}.")
        Tenant._print_json_response(response)

    @staticmethod
    def _print_json_response(response):
        try:
            json_response = response.json()
        except ValueError:
            json_response = "{}"
        print(json_response)


def write_to_namedtempfile(data, delete_tmp_files):
    temp = tempfile.NamedTemporaryFile(  # pylint: disable=consider-using-with
        prefix="keylime-", delete=delete_tmp_files
    )
    temp.write(data)
    temp.flush()
    return temp.name


def main(argv=sys.argv):  # pylint: disable=dangerous-default-value
    """[summary]

    Keyword Arguments:
        argv {[type]} -- [description] (default: {sys.argv})

    Raises:
        UserError: [description]
        UserError: [description]
        UserError: [description]
    """
    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument(
        "-c",
        "--command",
        action="store",
        dest="command",
        default="add",
        help="valid commands are add,delete,update,"
        "regstatus,cvstatus,status,reglist,cvlist,reactivate,"
        "regdelete,bulkinfo,addallowlist,showallowlist,deleteallowlist. defaults to add",
    )
    parser.add_argument(
        "-t", "--targethost", action="store", dest="agent_ip", help="the IP address of the host to provision"
    )
    parser.add_argument(
        "-tp", "--targetport", action="store", dest="agent_port", help="the Port of the host to provision"
    )
    parser.add_argument(
        "-r",
        "--registrarhost",
        action="store",
        dest="registrar_ip",
        help="the IP address of the registrar where to retrieve the agents data from.",
    )
    parser.add_argument(
        "-rp", "--registrarport", action="store", dest="registrar_port", help="the port of the registrar."
    )
    parser.add_argument(
        "--cv_targethost",
        action="store",
        default=None,
        dest="cv_agent_ip",
        help="the IP address of the host to provision that the verifier will use (optional).  Use only if different than argument to option -t/--targethost",
    )
    parser.add_argument("-v", "--cv", action="store", dest="verifier_ip", help="the IP address of the cloud verifier")
    parser.add_argument("-vp", "--cvport", action="store", dest="verifier_port", help="the port of the cloud verifier")
    parser.add_argument(
        "-vi", "--cvid", action="store", dest="verifier_id", help="the unique identifier of a cloud verifier"
    )
    parser.add_argument(
        "-nvc",
        "--no-verifier-check",
        action="store_false",
        dest="verifier_check",
        default=True,
        help="Disable the check to confirm if the agent is being processed by the specified verifier. Use only with -c/--command delete or reactivate",
    )
    parser.add_argument("-u", "--uuid", action="store", dest="agent_uuid", help="UUID for the agent to provision")
    parser.add_argument(
        "-f", "--file", action="store", default=None, help="Deliver the specified plaintext to the provisioned agent"
    )
    parser.add_argument(
        "--cert",
        action="store",
        dest="ca_dir",
        default=None,
        help='Create and deliver a certificate using a CA created by ca-util. Pass in the CA directory or use "default" to use the standard dir',
    )
    parser.add_argument(
        "-k", "--key", action="store", dest="keyfile", help="an intermedia key file produced by user_data_encrypt"
    )
    parser.add_argument(
        "-p",
        "--payload",
        action="store",
        default=None,
        help="Specify the encrypted payload to deliver with encrypted keys specified by -k",
    )
    parser.add_argument(
        "--include",
        action="store",
        dest="incl_dir",
        default=None,
        help="Include additional files in provided directory in certificate zip file.  Must be specified with --cert",
    )
    parser.add_argument(
        "--allowlist", action="store", dest="allowlist", default=None, help="Specify the file path of an allowlist"
    )
    parser.add_argument(
        "--signature-verification-key",
        "--sign_verification_key",
        action="append",
        dest="ima_sign_verification_keys",
        default=[],
        help="Specify an IMA file signature verification key",
    )
    parser.add_argument(
        "--signature-verification-key-sig",
        action="append",
        dest="ima_sign_verification_key_sigs",
        default=[],
        help="Specify the GPG signature file for an IMA file signature verification key; pair this option with --signature-verification-key",
    )
    parser.add_argument(
        "--signature-verification-key-sig-key",
        action="append",
        dest="ima_sign_verification_key_sig_keys",
        default=[],
        help="Specify the GPG public key file use to validate the --signature-verification-key-sig; pair this option with --signature-verification-key",
    )
    parser.add_argument(
        "--signature-verification-key-url",
        action="append",
        dest="ima_sign_verification_key_urls",
        default=[],
        help="Specify the URL for a remote IMA file signature verification key",
    )
    parser.add_argument(
        "--signature-verification-key-sig-url",
        action="append",
        dest="ima_sign_verification_key_sig_urls",
        default=[],
        help="Specify the URL for the remote GPG signature of a remote IMA file signature verification key; pair this option with --signature-verification-key-url",
    )
    parser.add_argument(
        "--signature-verification-key-sig-url-key",
        action="append",
        dest="ima_sign_verification_key_sig_url_keys",
        default=[],
        help="Specify the GPG public key file used to validate the --signature-verification-key-sig-url; pair this option with --signature-verification-key-url",
    )
    parser.add_argument(
        "--mb_refstate",
        action="store",
        dest="mb_refstate",
        default=None,
        help="Specify the location of a measure boot reference state (intended state)",
    )
    parser.add_argument(
        "--allowlist-checksum",
        action="store",
        dest="allowlist_checksum",
        default=None,
        help="Specify the SHA-256 checksum of an allowlist",
    )
    parser.add_argument(
        "--allowlist-sig",
        action="store",
        dest="allowlist_sig",
        default=None,
        help="Specify the GPG signature file of an allowlist",
    )
    parser.add_argument(
        "--allowlist-sig-key",
        action="store",
        dest="allowlist_sig_key",
        default=None,
        help="Specify the GPG public key file used to validate the --allowlist-sig or --allowlist-sig-url",
    )
    parser.add_argument(
        "--allowlist-url",
        action="store",
        dest="allowlist_url",
        default=None,
        help="Specify the URL of a remote allowlist",
    )
    parser.add_argument(
        "--allowlist-sig-url",
        action="store",
        dest="allowlist_sig_url",
        default=None,
        help="Specify the URL of the remote GPG signature file of an allowlist",
    )
    parser.add_argument(
        "--exclude",
        action="store",
        dest="ima_exclude",
        default=None,
        help="Specify the location of an IMA exclude list",
    )
    parser.add_argument(
        "--tpm_policy",
        action="store",
        dest="tpm_policy",
        default=None,
        help='Specify a TPM policy in JSON format. e.g., {"15":"0000000000000000000000000000000000000000"}',
    )
    parser.add_argument(
        "--verify",
        action="store_true",
        default=False,
        help="Block on cryptographically checked key derivation confirmation from the agent once it has been provisioned",
    )
    parser.add_argument("--allowlist-name", help="The name of allowlist to operate with")
    parser.add_argument(
        "--supported-version",
        default=None,
        action="store",
        dest="supported_version",
        help="API version that is supported by the agent. Detected automatically by default",
    )

    args = parser.parse_args(argv[1:])

    # Make sure argument dependencies are enforced
    if args.ima_exclude and not args.allowlist:
        parser.error("--exclude cannot be used without an --allowlist")
    if args.allowlist and args.allowlist_url:
        parser.error("--allowlist and --allowlist-url cannot be specified at the same time")
    if args.allowlist_url and not (args.allowlist_sig or args.allowlist_sig_url or args.allowlist_checksum):
        parser.error(
            "--allowlist-url must have either --allowlist-sig, --allowlist-sig-url or --allowlist-checksum to verifier integrity"
        )
    if args.allowlist_sig and not (args.allowlist_url or args.allowlist):
        parser.error("--allowlist-sig must have either --allowlist or --allowlist-url")
    if args.allowlist_sig_url and not (args.allowlist_url or args.allowlist):
        parser.error("--allowlist-sig-url must have either --allowlist or --allowlist-url")
    if args.allowlist_checksum and not (args.allowlist_url or args.allowlist):
        parser.error("--allowlist-checksum must have either --allowlist or --allowlist-url")
    if args.allowlist_sig and not args.allowlist_sig_key:
        parser.error("--allowlist-sig must also have --allowlist-sig-key")
    if args.allowlist_sig_url and not args.allowlist_sig_key:
        parser.error("--allowlist-sig-url must also have --allowlist-sig-key")
    if args.allowlist_sig_key and not (args.allowlist_sig or args.allowlist_sig_url):
        parser.error("--allowlist-sig-key must have either --allowlist-sig or --allowlist-sig-url")

    mytenant = Tenant()

    if args.agent_uuid is not None:
        mytenant.agent_uuid = args.agent_uuid
        # if the uuid is actually a public key, then hash it
        if mytenant.agent_uuid.startswith("-----BEGIN PUBLIC KEY-----"):
            mytenant.agent_uuid = hashlib.sha256(mytenant.agent_uuid).hexdigest()
        if not validators.valid_agent_id(mytenant.agent_uuid):
            raise UserError("The agent ID set via agent uuid parameter use invalid characters")
    else:
        logger.warning("Using default UUID d432fbb3-d2f1-4a97-9ef7-75bd81c00000")
        mytenant.agent_uuid = "d432fbb3-d2f1-4a97-9ef7-75bd81c00000"

    if args.verifier_id is not None:
        mytenant.verifier_id = args.verifier_id
    if args.verifier_ip is not None:
        mytenant.verifier_ip = args.verifier_ip
    if args.verifier_port is not None:
        mytenant.verifier_port = args.verifier_port

    if args.registrar_ip is not None:
        mytenant.registrar_ip = args.registrar_ip
    if args.registrar_port is not None:
        mytenant.registrar_port = args.registrar_port

    # we only need to fetch remote files if we are adding or updating
    if args.command in ["add", "update", "addallowlist"]:
        delete_tmp_files = logger.level > logging.DEBUG  # delete tmp files unless in DEBUG mode

        if args.allowlist_url:
            logger.info("Downloading Allowlist from %s", args.allowlist_url)
            response = requests.get(args.allowlist_url, timeout=mytenant.request_timeout, allow_redirects=False)
            if response.status_code == 200:
                args.allowlist = write_to_namedtempfile(response.content, delete_tmp_files)
                logger.debug("Allowlist temporarily saved in %s", args.allowlist)
            else:
                raise Exception(
                    f"Downloading allowlist ({args.allowlist_url}) failed with status code {response.status_code}!"
                )

        if args.allowlist_sig_url:
            logger.info("Downloading Allowlist signature from %s", args.allowlist_sig_url)
            response = requests.get(args.allowlist_sig_url, timeout=mytenant.request_timeout, allow_redirects=False)
            if response.status_code == 200:
                args.allowlist_sig = write_to_namedtempfile(response.content, delete_tmp_files)
                logger.debug("Allowlist signature temporarily saved in %s", args.allowlist_sig)
            else:
                raise Exception(
                    f"Downloading allowlist signature ({args.allowlist_sig_url}) failed with status code {response.status_code}!"
                )

        # verify all the local keys for which we have a signature file and a key to verify
        for i, key_file in enumerate(args.ima_sign_verification_keys):
            if len(args.ima_sign_verification_key_sigs) <= i:
                break
            keysig_file = args.ima_sign_verification_key_sigs[i]
            if len(args.ima_sign_verification_key_sig_keys) == 0:
                raise UserError(f"A gpg key is missing for key signature file '{keysig_file}'")

            gpg_key_file = args.ima_sign_verification_key_sig_keys[i]
            signing.verify_signature_from_file(gpg_key_file, key_file, keysig_file, "IMA file signing key")

            logger.info("Signature verification on %s was successful", key_file)

        # verify all the remote keys for which we have a signature URL and key to to verify
        # Append the downloaded key files to args.ima_sign_verification_keys
        for i, key_url in enumerate(args.ima_sign_verification_key_urls):

            logger.info("Downloading key from %s", key_url)
            response = requests.get(key_url, timeout=mytenant.request_timeout, allow_redirects=False)
            if response.status_code == 200:
                key_file = write_to_namedtempfile(response.content, delete_tmp_files)
                args.ima_sign_verification_keys.append(key_file)
                logger.debug("Key temporarily saved in %s", key_file)
            else:
                raise Exception(f"Downloading key ({key_url}) failed with status code {response.status_code}!")

            if len(args.ima_sign_verification_key_sig_urls) <= i:
                continue

            keysig_url = args.ima_sign_verification_key_sig_urls[i]

            if len(args.ima_sign_verification_key_sig_url_keys) == 0:
                raise UserError(f"A gpg key is missing for key signature URL '{keysig_url}'")

            logger.info("Downloading key signature from %s", keysig_url)
            response = requests.get(keysig_url, timeout=mytenant.request_timeout, allow_redirects=False)
            if response.status_code == 200:
                keysig_file = write_to_namedtempfile(response.content, delete_tmp_files)
                logger.debug("Key signature temporarily saved in %s", keysig_file)
            else:
                raise Exception(
                    f"Downloading key signature ({key_url}) failed with status code {response.status_code}!"
                )

            gpg_key_file = args.ima_sign_verification_key_sig_url_keys[i]
            signing.verify_signature_from_file(gpg_key_file, key_file, keysig_file, "IMA file signing key")
            logger.info("Signature verification on %s was successful", key_url)

    if args.command == "add":
        mytenant.init_add(vars(args))
        mytenant.preloop()
        mytenant.do_quote()
        mytenant.do_cv()
        if args.verify:
            mytenant.do_verify()
    elif args.command == "update":
        mytenant.init_add(vars(args))
        mytenant.do_cvdelete(args.verifier_check)
        mytenant.preloop()
        mytenant.do_quote()
        mytenant.do_cv()
        if args.verify:
            mytenant.do_verify()
    elif args.command == "delete":
        mytenant.do_cvdelete(args.verifier_check)
    elif args.command == "status":
        mytenant.do_status()
    elif args.command == "cvstatus":
        mytenant.do_cvstatus()
    elif args.command == "bulkinfo":
        mytenant.do_cvbulkinfo()
    elif args.command == "cvlist":
        mytenant.do_cvlist()
    elif args.command == "reactivate":
        mytenant.do_cvreactivate(args.verifier_check)
    elif args.command == "regstatus":
        mytenant.do_regstatus()
    elif args.command == "reglist":
        mytenant.do_reglist()
    elif args.command == "regdelete":
        mytenant.do_regdelete()
    elif args.command == "addallowlist":
        mytenant.do_add_allowlist(vars(args))
    elif args.command == "showallowlist":
        mytenant.do_show_allowlist(args.allowlist_name)
    elif args.command == "deleteallowlist":
        mytenant.do_delete_allowlist(args.allowlist_name)
    else:
        raise UserError(f"Invalid command specified: {args.command}")



import os

import alembic.config

from keylime import keylime_logging


def main():
    apply(None)


def apply(db_name):
    # set a conservative general umask
    os.umask(0o077)

    here = os.path.dirname(os.path.abspath(__file__))

    # the config file for alembic is in the migrations directory
    alembic_args = ["-c", os.path.join(here, "..", "migrations", "alembic.ini")]

    # if we are restricting it to a single db, add that to the custom args (-x)
    if db_name:
        alembic_args.extend(["-x", "db=" + db_name])

    alembic_args.extend(["upgrade", "head"])

    alembic.config.main(argv=alembic_args)


if __name__ == "__main__":
    logger = keylime_logging.init_logging("migrations")
    try:
        main()
    except Exception as e:
        logger.exception(e)



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2021 Angelo Ruocco - IBM Research Lab Zurich
"""

import random
import unittest

from keylime.common import retry


def rand_base():
    return random.uniform(-99, 99)


def rand_base_proper():
    return random.uniform(1, 99)


def rand_ntries():
    return random.randint(1, 99)


def rand_exp():
    return random.choice([True, False])


class RetryInterval_Test(unittest.TestCase):
    def test_general(self):
        self.assertTrue(retry.retry_time(rand_exp(), rand_base(), rand_ntries(), None) >= 0)

    def test_linear(self):
        b = rand_base()
        self.assertEqual(retry.retry_time(False, b, rand_ntries, None), abs(b))

    def test_exponential(self):
        b0 = rand_base_proper()
        b1 = random.random()
        n = rand_ntries()
        self.assertEqual(retry.retry_time(True, b0, n, None), b0**n)
        self.assertEqual(retry.retry_time(True, b1, n, None), abs(b1))


if __name__ == "__main__":
    unittest.main()



"""Add ima_sign_verification_keys column

Revision ID: a09a40352c32
Revises: eeb702f77d7d
Create Date: 2020-11-18 10:37:58.244212

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a7a64155ab3a"
down_revision = "8da20383f6e1"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("ima_sign_verification_keys", sa.String(1000)))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "ima_sign_verification_keys")



import sys
import threading

from keylime.common.algorithms import Hash
from keylime.ima.ast import get_FF_HASH, get_START_HASH
from keylime.ima.file_signatures import ImaKeyrings

if sys.version_info >= (3, 7):
    from dataclasses import dataclass
else:
    from keylime.backport_dataclasses import dataclass


@dataclass
class TPMClockInfo:
    clock: int
    resetcount: int
    restartcount: int
    safe: int

    @classmethod
    def from_dict(cls, data) -> "TPMClockInfo":
        dclki = {}
        if "clockInfo" in data:
            dclki = data["clockInfo"]

        if "clock" in data:
            dclki = data

        return cls(
            clock=dclki.get("clock", 0),
            resetcount=dclki.get("resetCount", 0),
            restartcount=dclki.get("restartCount", 0),
            safe=dclki.get("safe", 1),
        )

    def to_dict(self):
        data = {}
        data["clock"] = self.clock
        data["resetCount"] = self.resetcount
        data["restartCount"] = self.restartcount
        data["safe"] = self.safe
        return data


class TPMState:
    """TPMState models the state of the TPM's PCRs"""

    def __init__(self):
        """constructor"""
        self.pcrs = {}
        self.hash_alg = {}  # Record the hash algorithm that a given PCR uses
        for pcr_num in range(0, 24):
            self.reset_pcr(pcr_num)

    def init_pcr(self, pcr_num, hash_alg):
        """Initializes a PCR"""
        if pcr_num not in self.hash_alg:
            self.hash_alg[pcr_num] = hash_alg

        if self.pcrs[pcr_num] is None:
            if 17 <= pcr_num <= 23:
                self.pcrs[pcr_num] = get_FF_HASH(self.hash_alg[pcr_num])
            else:
                self.pcrs[pcr_num] = get_START_HASH(self.hash_alg[pcr_num])

    def reset_pcr(self, pcr_num):
        """Reset a specific PCR"""
        self.pcrs[pcr_num] = None

    def get_pcr(self, pcr_num):
        """
        Get the state of a PCR.

        :returns: PCR value or None if not initialized
        """
        return self.pcrs.get(pcr_num, None)

    def used_pcr(self, pcr_num):
        """Check if a PCR was actually requested at least once"""
        return (pcr_num in self.hash_alg) and (self.pcrs[pcr_num] is not None)

    def set_pcr(self, pcr_num, pcr_value):
        """Set the value of a PCR"""
        self.pcrs[pcr_num] = pcr_value


class AgentAttestState:
    """AgentAttestState is used to support incremental attestation"""

    def __init__(self, agent_id):
        """constructor"""

        self.agent_id = agent_id
        self.next_ima_ml_entry = 0
        self.set_boottime(0)
        self.tpm_clockinfo = TPMClockInfo(clock=0, resetcount=0, restartcount=0, safe=1)

        self.tpm_state = TPMState()
        self.ima_pcrs = set()

        self.ima_keyrings = ImaKeyrings()

        self.reset_ima_attestation()

        self.ima_dm_state = None

    def get_agent_id(self):
        """Get the agent_id"""
        return self.agent_id

    def reset_ima_attestation(self):
        """Reset the IMA attestation state to start over with 1st entry
        ad start over with learning the keys"""
        self.next_ima_ml_entry = 0
        for pcr_num in self.ima_pcrs:
            self.tpm_state.reset_pcr(pcr_num)
        self.set_boottime(0)
        self.ima_keyrings = ImaKeyrings()

    def update_ima_attestation(self, pcr_num, pcr_value, num_ml_entries):
        """Update the attestation by remembering the new PCR value and the
        number of lines that were successfully processed."""
        self.ima_pcrs.add(pcr_num)
        self.tpm_state.set_pcr(pcr_num, pcr_value)
        self.next_ima_ml_entry += num_ml_entries

    def get_ima_pcrs(self):
        """Return a dict with the IMA pcrs"""
        ima_pcrs_dict = {}
        for pcr_num in self.ima_pcrs:
            # Only output IMA PCRs that were used at least once
            if self.tpm_state.used_pcr(pcr_num):
                ima_pcrs_dict[pcr_num] = self.tpm_state.get_pcr(pcr_num)
        return ima_pcrs_dict

    def set_ima_pcrs(self, ima_pcrs_dict):
        """Set the values of the given ima_pcrs dict in the tpm_state"""
        for pcr_num, pcr_value in ima_pcrs_dict.items():
            self.tpm_state.set_pcr(pcr_num, pcr_value)
        self.ima_pcrs = set(ima_pcrs_dict.keys())

    def get_next_ima_ml_entry(self):
        """Return the next IMA measurement list entry we want to request from agent"""
        return self.next_ima_ml_entry

    def set_next_ima_ml_entry(self, next_ima_ml_entry):
        """Set the value of the next_ima_ml_entry field"""
        self.next_ima_ml_entry = next_ima_ml_entry

    def get_pcr_state(self, pcr_num, hash_alg=Hash.SHA1):
        """Return the PCR state of the given PCR"""
        if not self.tpm_state.used_pcr(pcr_num):
            self.tpm_state.init_pcr(pcr_num, hash_alg)
        return self.tpm_state.get_pcr(pcr_num)

    def get_boottime(self):
        """Return the boottime of the system"""
        return self.boottime

    def set_boottime(self, boottime):
        """Set the boottime of the system"""
        self.boottime = boottime

    def get_tpm_clockinfo(self):
        """Return the clock info extracted from a TPM quote"""
        return self.tpm_clockinfo

    def set_tpm_clockinfo(self, tpm_clockinfo):
        """Set the clock info with information extracted from a TPM quote"""
        self.tpm_clockinfo = tpm_clockinfo

    def is_expected_boottime(self, boottime):
        """Check whether the given boottime is the expected boottime"""
        return self.boottime == boottime

    def set_ima_keyrings(self, ima_keyrings):
        """Set the ImaKeyrings object"""
        self.ima_keyrings = ima_keyrings

    def get_ima_keyrings(self):
        """Get the ImaKeyrings object"""
        return self.ima_keyrings

    def get_ima_dm_state(self):
        """Get encoded state of the DmValidator"""
        return self.ima_dm_state

    def set_ima_dm_state(self, state):
        self.ima_dm_state = state


class AgentAttestStates:
    """AgentAttestStates administers a map of AgentAttestState's indexed by agent_id"""

    instance = None

    @staticmethod
    def get_instance():
        """Create and return a singleton AgentAttestState"""
        if not AgentAttestStates.instance:
            AgentAttestStates.instance = AgentAttestStates()
        return AgentAttestStates.instance

    def __init__(self):
        """constructor"""
        self.map_lock = threading.Lock()
        self.map = {}

    def get_by_agent_id(self, agent_id):
        """Get an agent's state given its id"""

        with self.map_lock:
            agentAttestState = self.map.get(agent_id)
            if not agentAttestState:
                agentAttestState = AgentAttestState(agent_id)
                self.map[agent_id] = agentAttestState

        return agentAttestState

    def delete_by_agent_id(self, agent_id):
        """Delete an agent's state given its id"""

        with self.map_lock:
            try:
                del self.map[agent_id]
            except KeyError:
                pass

    def add(self, agent_id, boottime, ima_pcrs_dict, next_ima_ml_entry, learned_ima_keyrings):
        """Add or replace an existing AgentAttestState initialized with the given values"""
        agentAttestState = self.get_by_agent_id(agent_id)
        agentAttestState.set_boottime(boottime)
        agentAttestState.set_ima_pcrs(ima_pcrs_dict)
        agentAttestState.set_next_ima_ml_entry(next_ima_ml_entry)
        agentAttestState.set_ima_keyrings(ImaKeyrings.from_json(learned_ima_keyrings))



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2021 Red Hat, Inc.
"""


import os
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

from keylime import ca_util

# Useful constants for the test
PACKAGE_ROOT = Path(__file__).parents[1]
CODE_ROOT = f"{PACKAGE_ROOT}/keylime/"

# Custom imports
sys.path.insert(0, CODE_ROOT)


class CA_Util_Test(unittest.TestCase):
    def test_load_cert_by_path(self):
        curdir = os.path.dirname(os.path.abspath(__file__))
        cert_path = os.path.join(curdir, "data", "ca", "cacert.crt")
        cert = ca_util.load_cert_by_path(cert_path)

        self.assertEqual(cert.serial_number, 1)

    def test_get_crl_distpoint(self):
        curdir = os.path.dirname(os.path.abspath(__file__))
        cert_path = os.path.join(curdir, "data", "ca", "cacert.crt")

        crl_distpoint = ca_util.get_crl_distpoint(cert_path)
        self.assertEqual(crl_distpoint, "http://localhost/crl.pem")

    def test_ca_util(self):
        ca_util.setpassword("42")

        try:
            # Create directory to be our working dir.
            working_dir = tempfile.mkdtemp()

            # cmd_init()
            ca_util.cmd_init(working_dir)

            # cmd_mkcert()
            ca_util.cmd_mkcert(working_dir, "foo bar")

            # cmd_certpkg()
            ca_util.cmd_certpkg(working_dir, "foo bar")

            # cmd_revoke()
            ca_util.cmd_revoke(working_dir, "foo bar")

            # cmd_regencrl()
            ca_util.cmd_regencrl(working_dir)
        except Exception as e:
            self.fail(e)
        finally:
            # Remove temporary directory.
            shutil.rmtree(working_dir)



from keylime import ca_util, keylime_logging

logger = keylime_logging.init_logging("ca-util")


def main():
    ca_util.main()


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        logger.exception(e)



"""add_last_received_quote

Revision ID: a09cc94177f0
Revises: 4089e1c79be9
Create Date: 2022-08-30 12:16:37.506940

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a09cc94177f0"
down_revision = "4089e1c79be9"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("last_received_quote", sa.Integer(), nullable=True))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "last_received_quote")



import datetime

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding, load_pem_private_key
from cryptography.x509.oid import NameOID

from keylime import config


def mk_cert_valid(cert_req, days=365):
    """
    Make a cert valid from now and til 'days' from now.
    Args:
       cert_req -- cryptography.x509.base.CertificateBuilder
       days -- number of days cert is valid for from now.
    Returns: updated cryptography.x509.base.CertificateBuilder
    """

    one_day = datetime.timedelta(1, 0, 0)
    today = datetime.datetime.now(datetime.timezone.utc)
    cert_req = cert_req.not_valid_before(today)
    cert_req = cert_req.not_valid_after(today + (one_day * days))
    return cert_req


def mk_name(common_name):
    return x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, config.get("ca", "cert_country")),
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, config.get("ca", "cert_state")),
            x509.NameAttribute(NameOID.LOCALITY_NAME, config.get("ca", "cert_locality")),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, config.get("ca", "cert_organization")),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, config.get("ca", "cert_org_unit")),
        ]
    )


def mk_request(bits, common_name):
    """
    Create a X509 request with the given number of bits in they key.
    Args:
      bits -- number of RSA key bits
      common_name -- common name in the request
    Returns a X509 request and the private key
    """

    privkey = rsa.generate_private_key(
        public_exponent=65537,
        key_size=bits,
        backend=default_backend(),
    )

    cert_req = x509.CertificateBuilder()

    subject = mk_name(common_name)
    cert_req = cert_req.subject_name(subject)
    return cert_req, privkey


def mk_cacert(name=None):
    """
    Make a CA certificate.
    Returns the certificate, private key and public key.
    """

    if name is None:
        name = config.get("ca", "cert_ca_name")
    cert_req, privkey = mk_request(config.getint("ca", "cert_bits"), name)

    pubkey = privkey.public_key()
    cert_req = cert_req.public_key(pubkey)

    cert_req = cert_req.serial_number(1)
    cert_req = mk_cert_valid(cert_req, config.getint("ca", "cert_ca_lifetime"))
    cert_req = cert_req.issuer_name(mk_name(name))

    # Extensions.
    extensions = [
        # Basic Constraints.
        x509.BasicConstraints(ca=True, path_length=None),
        # Subject Key Identifier.
        x509.SubjectKeyIdentifier.from_public_key(pubkey),
        # CRL Distribution Points.
        x509.CRLDistributionPoints(
            [
                x509.DistributionPoint(
                    full_name=[
                        x509.UniformResourceIdentifier("http://localhost/crl.pem"),
                    ],
                    relative_name=None,
                    reasons=None,
                    crl_issuer=None,
                ),
            ]
        ),
        # Key Usage.
        x509.KeyUsage(
            key_cert_sign=True,
            crl_sign=True,
            digital_signature=False,
            content_commitment=False,
            key_encipherment=False,
            data_encipherment=False,
            key_agreement=False,
            encipher_only=False,
            decipher_only=False,
        ),
    ]

    for ext in extensions:
        cert_req = cert_req.add_extension(ext, critical=False)

    cert = cert_req.sign(
        private_key=privkey,
        algorithm=hashes.SHA256(),
        backend=default_backend(),
    )

    return cert, privkey, pubkey


def mk_signed_cert(cacert, ca_privkey, name, serialnum):
    """
    Create a CA cert + server cert + server private key.
    """

    cert_req, privkey = mk_request(config.getint("ca", "cert_bits"), common_name=name)
    pubkey = privkey.public_key()
    cert_req = cert_req.public_key(pubkey)

    cert_req = cert_req.serial_number(serialnum)
    cert_req = mk_cert_valid(cert_req)
    cert_req = cert_req.issuer_name(cacert.issuer)

    # Extensions.
    extensions = [
        # OID 2.16.840.1.113730.1.13 is Netscape Comment.
        # http://oid-info.com/get/2.16.840.1.113730.1.13
        x509.UnrecognizedExtension(
            oid=x509.ObjectIdentifier("2.16.840.1.113730.1.13"),
            value=b"SSL Server",
        ),
        # Subject Alternative Name.
        x509.SubjectAlternativeName([x509.DNSName(name)]),
        # CRL Distribution Points.
        x509.CRLDistributionPoints(
            [
                x509.DistributionPoint(
                    full_name=[
                        x509.UniformResourceIdentifier("http://localhost/crl.pem"),
                    ],
                    relative_name=None,
                    reasons=None,
                    crl_issuer=None,
                ),
            ]
        ),
    ]

    for ext in extensions:
        cert_req = cert_req.add_extension(ext, critical=False)

    cert = cert_req.sign(
        private_key=ca_privkey,
        algorithm=hashes.SHA256(),
        backend=default_backend(),
    )
    return cert, privkey


def gencrl(serials, cert: str, ca_pk: str):
    ca_cert = x509.load_pem_x509_certificate(cert.encode())
    priv_key = load_pem_private_key(ca_pk.encode(), None, backend=default_backend())
    date_now = datetime.datetime.now(datetime.timezone.utc)

    builder = x509.CertificateRevocationListBuilder()
    builder = builder.issuer_name(ca_cert.issuer)
    builder = builder.last_update(date_now)
    builder = builder.next_update(date_now)

    for serial in serials:
        cert = x509.RevokedCertificateBuilder()
        cert = cert.serial_number(int(serial))
        cert = cert.revocation_date(date_now)
        builder = builder.add_revoked_certificate(cert.build())

    crl = builder.sign(private_key=priv_key, algorithm=hashes.SHA256(), backend=default_backend())
    return crl.public_bytes(encoding=Encoding.DER)



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2020 Luke Hinds (lhinds@redhat.com), Red Hat, Inc.
"""

import unittest

from sqlalchemy import create_engine

from keylime.db.keylime_db import SessionManager
from keylime.db.registrar_db import RegistrarMain

# BEGIN TEST DATA

test_data = {
    "agent_id": "d432fbb3-d2f1-4a97-9ef7-75bd81c00000",
    "ek_tpm": """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0dLxdAABVJO6qxamjCMh
yhWZgiFHZHnPEe0tMFyK3fNVr/w8lX9r+QOLxLmkT0IdgsEYtGZGefbD+qQl4O1s
k25823Xzu5tEF8966rTdkfsv8CRrNaBLwWlnt/n+qjIoU3xZJMmR+mFfqTc3a6zV
mPOYJstFtM8r4b9HPCUq6Mte/J3Wx4FxI9R4UrCUyiAeH++0QapIxuEGsVIYs92n
GyvFQYBZFRU6cIt33iaqTrRCICJp+YblMnw54YJGAH2vTVQf6/fLAnQt5L1UfmTy
R/ZA6advx8soekSBOIAW7XmV8Xp9mSquIHZdSXMJlcn/B35PU3BdkUtIYm5JuGGt
PQIDAQAB
-----END PUBLIC KEY-----""",
    "aik_tpm": """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0dLxdAABVJO6qxamjCMh
yhWZgiFHZHnPEe0tMFyK3fNVr/w8lX9r+QOLxLmkT0IdgsEYtGZGefbD+qQl4O1s
k25823Xzu5tEF8966rTdkfsv8CRrNaBLwWlnt/n+qjIoU3xZJMmR+mFfqTc3a6zV
mPOYJstFtM8r4b9HPCUq6Mte/J3Wx4FxI9R4UrCUyiAeH++0QapIxuEGsVIYs92n
GyvFQYBZFRU6cIt33iaqTrRCICJp+YblMnw54YJGAH2vTVQf6/fLAnQt5L1UfmTy
R/ZA6advx8soekSBOIAW7XmV8Xp9mSquIHZdSXMJlcn/B35PU3BdkUtIYm5JuGGt
PQIDAQAB
-----END PUBLIC KEY-----""",
    "ekcert": """-----BEGIN CERTIFICATE-----
MIIEnzCCA4egAwIBAgIEMV64bDANBgkqhkiG9w0BAQUFADBtMQswCQYDVQQGEwJE
RTEQMA4GA1UECBMHQmF2YXJpYTEhMB8GA1UEChMYSW5maW5lb24gVGVjaG5vbG9n
aWVzIEFHMQwwCgYDVQQLEwNBSU0xGzAZBgNVBAMTEklGWCBUUE0gRUsgUm9vdCBD
QTAeFw0wNTEwMjAxMzQ3NDNaFw0yNTEwMjAxMzQ3NDNaMHcxCzAJBgNVBAYTAkRF
MQ8wDQYDVQQIEwZTYXhvbnkxITAfBgNVBAoTGEluZmluZW9uIFRlY2hub2xvZ2ll
cyBBRzEMMAoGA1UECxMDQUlNMSYwJAYDVQQDEx1JRlggVFBNIEVLIEludGVybWVk
aWF0ZSBDQSAwMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALftPhYN
t4rE+JnU/XOPICbOBLvfo6iA7nuq7zf4DzsAWBdsZEdFJQfaK331ihG3IpQnlQ2i
YtDim289265f0J4OkPFpKeFU27CsfozVaNUm6UR/uzwA8ncxFc3iZLRMRNLru/Al
VG053ULVDQMVx2iwwbBSAYO9pGiGbk1iMmuZaSErMdb9v0KRUyZM7yABiyDlM3cz
UQX5vLWV0uWqxdGoHwNva5u3ynP9UxPTZWHZOHE6+14rMzpobs6Ww2RR8BgF96rh
4rRAZEl8BXhwiQq4STvUXkfvdpWH4lzsGcDDtrB6Nt3KvVNvsKz+b07Dk+Xzt+EH
NTf3Byk2HlvX+scCAwEAAaOCATswggE3MB0GA1UdDgQWBBQ4k8292HPEIzMV4bE7
qWoNI8wQxzAOBgNVHQ8BAf8EBAMCAgQwEgYDVR0TAQH/BAgwBgEB/wIBADBYBgNV
HSABAf8ETjBMMEoGC2CGSAGG+EUBBy8BMDswOQYIKwYBBQUHAgEWLWh0dHA6Ly93
d3cudmVyaXNpZ24uY29tL3JlcG9zaXRvcnkvaW5kZXguaHRtbDCBlwYDVR0jBIGP
MIGMgBRW65FEhWPWcrOu1EWWC/eUDlRCpqFxpG8wbTELMAkGA1UEBhMCREUxEDAO
BgNVBAgTB0JhdmFyaWExITAfBgNVBAoTGEluZmluZW9uIFRlY2hub2xvZ2llcyBB
RzEMMAoGA1UECxMDQUlNMRswGQYDVQQDExJJRlggVFBNIEVLIFJvb3QgQ0GCAQMw
DQYJKoZIhvcNAQEFBQADggEBABJ1+Ap3rNlxZ0FW0aIgdzktbNHlvXWNxFdYIBbM
OKjmbOos0Y4O60eKPu259XmMItCUmtbzF3oKYXq6ybARUT2Lm+JsseMF5VgikSlU
BJALqpKVjwAds81OtmnIQe2LSu4xcTSavpsL4f52cUAu/maMhtSgN9mq5roYptq9
DnSSDZrX4uYiMPl//rBaNDBflhJ727j8xo9CCohF3yQUoQm7coUgbRMzyO64yMIO
3fhb+Vuc7sNwrMOz3VJN14C3JMoGgXy0c57IP/kD5zGRvljKEvrRC2I147+fPeLS
DueRMS6lblvRKiZgmGAg7YaKOkOaEmVDMQ+fTo2Po7hI5wc=
-----END CERTIFICATE-----""",
    "virtual": 0,
    "active": int(False),
    "key": "R0xKMU5maWVEaHVxQ0poZUZhWm9yRVkyRHZZUkVIMFA=",
    "provider_keys": {},
    "regcount": 1,
    "ip": "127.0.0.1",
    "port": 9002,
}

agent_id = "d432fbb3-d2f1-4a97-9ef7-75bd81c00000"

# END TEST DATA


class TestRegistrarDB(unittest.TestCase):
    def setUp(self):
        self.engine = create_engine("sqlite://")
        RegistrarMain.metadata.create_all(self.engine, checkfirst=True)
        self.session = SessionManager().make_session(self.engine)
        self.populate_agent()

    def populate_agent(self):
        self.session.add(RegistrarMain(**test_data))
        self.session.commit()

    def test_add_agent(self):
        agent = self.session.query(RegistrarMain).filter_by(agent_id=agent_id).first()
        self.assertEqual(
            agent.ek_tpm,
            """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0dLxdAABVJO6qxamjCMh
yhWZgiFHZHnPEe0tMFyK3fNVr/w8lX9r+QOLxLmkT0IdgsEYtGZGefbD+qQl4O1s
k25823Xzu5tEF8966rTdkfsv8CRrNaBLwWlnt/n+qjIoU3xZJMmR+mFfqTc3a6zV
mPOYJstFtM8r4b9HPCUq6Mte/J3Wx4FxI9R4UrCUyiAeH++0QapIxuEGsVIYs92n
GyvFQYBZFRU6cIt33iaqTrRCICJp+YblMnw54YJGAH2vTVQf6/fLAnQt5L1UfmTy
R/ZA6advx8soekSBOIAW7XmV8Xp9mSquIHZdSXMJlcn/B35PU3BdkUtIYm5JuGGt
PQIDAQAB
-----END PUBLIC KEY-----""",
        )
        self.assertEqual(
            agent.aik_tpm,
            """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0dLxdAABVJO6qxamjCMh
yhWZgiFHZHnPEe0tMFyK3fNVr/w8lX9r+QOLxLmkT0IdgsEYtGZGefbD+qQl4O1s
k25823Xzu5tEF8966rTdkfsv8CRrNaBLwWlnt/n+qjIoU3xZJMmR+mFfqTc3a6zV
mPOYJstFtM8r4b9HPCUq6Mte/J3Wx4FxI9R4UrCUyiAeH++0QapIxuEGsVIYs92n
GyvFQYBZFRU6cIt33iaqTrRCICJp+YblMnw54YJGAH2vTVQf6/fLAnQt5L1UfmTy
R/ZA6advx8soekSBOIAW7XmV8Xp9mSquIHZdSXMJlcn/B35PU3BdkUtIYm5JuGGt
PQIDAQAB
-----END PUBLIC KEY-----""",
        )
        self.assertEqual(
            agent.ekcert,
            """-----BEGIN CERTIFICATE-----
MIIEnzCCA4egAwIBAgIEMV64bDANBgkqhkiG9w0BAQUFADBtMQswCQYDVQQGEwJE
RTEQMA4GA1UECBMHQmF2YXJpYTEhMB8GA1UEChMYSW5maW5lb24gVGVjaG5vbG9n
aWVzIEFHMQwwCgYDVQQLEwNBSU0xGzAZBgNVBAMTEklGWCBUUE0gRUsgUm9vdCBD
QTAeFw0wNTEwMjAxMzQ3NDNaFw0yNTEwMjAxMzQ3NDNaMHcxCzAJBgNVBAYTAkRF
MQ8wDQYDVQQIEwZTYXhvbnkxITAfBgNVBAoTGEluZmluZW9uIFRlY2hub2xvZ2ll
cyBBRzEMMAoGA1UECxMDQUlNMSYwJAYDVQQDEx1JRlggVFBNIEVLIEludGVybWVk
aWF0ZSBDQSAwMTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALftPhYN
t4rE+JnU/XOPICbOBLvfo6iA7nuq7zf4DzsAWBdsZEdFJQfaK331ihG3IpQnlQ2i
YtDim289265f0J4OkPFpKeFU27CsfozVaNUm6UR/uzwA8ncxFc3iZLRMRNLru/Al
VG053ULVDQMVx2iwwbBSAYO9pGiGbk1iMmuZaSErMdb9v0KRUyZM7yABiyDlM3cz
UQX5vLWV0uWqxdGoHwNva5u3ynP9UxPTZWHZOHE6+14rMzpobs6Ww2RR8BgF96rh
4rRAZEl8BXhwiQq4STvUXkfvdpWH4lzsGcDDtrB6Nt3KvVNvsKz+b07Dk+Xzt+EH
NTf3Byk2HlvX+scCAwEAAaOCATswggE3MB0GA1UdDgQWBBQ4k8292HPEIzMV4bE7
qWoNI8wQxzAOBgNVHQ8BAf8EBAMCAgQwEgYDVR0TAQH/BAgwBgEB/wIBADBYBgNV
HSABAf8ETjBMMEoGC2CGSAGG+EUBBy8BMDswOQYIKwYBBQUHAgEWLWh0dHA6Ly93
d3cudmVyaXNpZ24uY29tL3JlcG9zaXRvcnkvaW5kZXguaHRtbDCBlwYDVR0jBIGP
MIGMgBRW65FEhWPWcrOu1EWWC/eUDlRCpqFxpG8wbTELMAkGA1UEBhMCREUxEDAO
BgNVBAgTB0JhdmFyaWExITAfBgNVBAoTGEluZmluZW9uIFRlY2hub2xvZ2llcyBB
RzEMMAoGA1UECxMDQUlNMRswGQYDVQQDExJJRlggVFBNIEVLIFJvb3QgQ0GCAQMw
DQYJKoZIhvcNAQEFBQADggEBABJ1+Ap3rNlxZ0FW0aIgdzktbNHlvXWNxFdYIBbM
OKjmbOos0Y4O60eKPu259XmMItCUmtbzF3oKYXq6ybARUT2Lm+JsseMF5VgikSlU
BJALqpKVjwAds81OtmnIQe2LSu4xcTSavpsL4f52cUAu/maMhtSgN9mq5roYptq9
DnSSDZrX4uYiMPl//rBaNDBflhJ727j8xo9CCohF3yQUoQm7coUgbRMzyO64yMIO
3fhb+Vuc7sNwrMOz3VJN14C3JMoGgXy0c57IP/kD5zGRvljKEvrRC2I147+fPeLS
DueRMS6lblvRKiZgmGAg7YaKOkOaEmVDMQ+fTo2Po7hI5wc=
-----END CERTIFICATE-----""",
        )
        self.assertEqual(agent.virtual, 0)
        self.assertEqual(agent.active, int(False))
        self.assertEqual(agent.key, "R0xKMU5maWVEaHVxQ0poZUZhWm9yRVkyRHZZUkVIMFA=")
        self.assertEqual(agent.provider_keys, {})
        self.assertEqual(agent.regcount, 1)
        self.assertEqual(agent.ip, "127.0.0.1")
        self.assertEqual(agent.port, 9002)

    def test_delete_agent(self):
        agent = self.session.query(RegistrarMain).filter_by(agent_id=agent_id).first()
        self.session.query(RegistrarMain).filter_by(agent_id=agent_id).delete()
        self.session.commit()
        agent = self.session.query(RegistrarMain).filter_by(agent_id=agent_id).first()
        self.assertIsNone(agent)

    def tearDown(self):
        self.session.close()



import ast
import base64
import time

from keylime import config, crypto, json, keylime_logging
from keylime.agentstates import AgentAttestStates, TPMClockInfo
from keylime.common import algorithms, validators
from keylime.failure import Component, Failure
from keylime.ima import file_signatures
from keylime.tpm.tpm_abstract import TPM_Utilities
from keylime.tpm.tpm_main import tpm

# setup logging
logger = keylime_logging.init_logging("cloudverifier_common")

GLOBAL_TPM_INSTANCE = None
DEFAULT_VERIFIER_ID = "default"


def get_tpm_instance():
    global GLOBAL_TPM_INSTANCE
    if GLOBAL_TPM_INSTANCE is None:
        GLOBAL_TPM_INSTANCE = tpm()
    return GLOBAL_TPM_INSTANCE


def get_AgentAttestStates():
    return AgentAttestStates.get_instance()


def process_quote_response(agent, ima_policy, json_response, agentAttestState) -> Failure:
    """Validates the response from the Cloud agent.

    This method invokes an Registrar Server call to register, and then check the quote.
    """
    failure = Failure(Component.QUOTE_VALIDATION)
    received_public_key = None
    quote = None
    agent_id = None
    # in case of failure in response content do not continue
    try:
        agent_id = agent["agent_id"]
        received_public_key = json_response.get("pubkey", None)
        quote = json_response["quote"]

        ima_measurement_list = json_response.get("ima_measurement_list", None)
        ima_measurement_list_entry = json_response.get("ima_measurement_list_entry", 0)
        mb_measurement_list = json_response.get("mb_measurement_list", None)
        boottime = json_response.get("boottime", 0)

        logger.debug(
            "received data for agent %s, quote: %s, nonce: %s, public key(b64): %s, ima_measurement_list: %s, ima_measurement_list_entry: %s, measured_boot_log: %s, boottime: %s",
            agent_id,
            quote,
            agent["nonce"],
            base64.b64encode(str(received_public_key).encode("ascii")).decode("ascii"),
            (ima_measurement_list is not None),
            ima_measurement_list_entry,
            (mb_measurement_list is not None),
            boottime,
        )

    except Exception as e:
        failure.add_event("invalid_data", {"message": "parsing agent get quote respone failed", "data": e}, False)
        return failure

    # TODO: Are those separate failures?
    if not isinstance(ima_measurement_list_entry, int):
        raise Exception("ima_measurement_list_entry parameter must be an integer")

    if not isinstance(boottime, int):
        raise Exception("boottime parameter must be an integer")

    # if no public key provided, then ensure we have cached it
    if received_public_key is None:
        if agent.get("public_key", "") == "" or agent.get("b64_encrypted_V", "") == "":
            logger.error("agent %s did not provide public key and no key or encrypted_v was cached at CV", agent_id)
            failure.add_event(
                "no_pubkey",
                {
                    "message": "agent did not provide public key and no key or encrypted_v was cached at CV",
                    "data": received_public_key,
                },
                False,
            )
            return failure
        agent["provide_V"] = False
        received_public_key = agent["public_key"]

    hash_alg = json_response.get("hash_alg")
    enc_alg = json_response.get("enc_alg")
    sign_alg = json_response.get("sign_alg")

    # Update chosen tpm and algorithms
    agent["hash_alg"] = hash_alg
    agent["enc_alg"] = enc_alg
    agent["sign_alg"] = sign_alg

    # Ensure hash_alg is in accept_tpm_hash_alg list
    if not algorithms.is_accepted(hash_alg, agent["accept_tpm_hash_algs"]) or not algorithms.Hash.is_recognized(
        hash_alg
    ):
        logger.error("TPM Quote for agent %s is using an unaccepted hash algorithm: %s", agent_id, hash_alg)
        failure.add_event(
            "invalid_hash_alg",
            {"message": f"TPM Quote is using an unaccepted hash algorithm: {hash_alg}", "data": hash_alg},
            False,
        )
        return failure

    # Ensure enc_alg is in accept_tpm_encryption_algs list
    if not algorithms.is_accepted(enc_alg, agent["accept_tpm_encryption_algs"]):
        logger.error("TPM Quote for agent %s is using an unaccepted encryption algorithm: %s", agent_id, enc_alg)
        failure.add_event(
            "invalid_enc_alg",
            {"message": f"TPM Quote is using an unaccepted encryption algorithm: {enc_alg}", "data": enc_alg},
            False,
        )
        return failure

    # Ensure sign_alg is in accept_tpm_encryption_algs list
    if not algorithms.is_accepted(sign_alg, agent["accept_tpm_signing_algs"]):
        logger.error("TPM Quote for agent %s is using an unaccepted signing algorithm: %s", agent_id, sign_alg)
        failure.add_event(
            "invalid_sign_alg",
            {"message": f"TPM Quote is using an unaccepted signing algorithm: {sign_alg}", "data": {sign_alg}},
            False,
        )
        return failure

    if ima_measurement_list_entry == 0:
        agentAttestState.reset_ima_attestation()
    elif ima_measurement_list_entry != agentAttestState.get_next_ima_ml_entry():
        # If we requested a particular entry number then the agent must return either
        # starting at 0 (handled above) or with the requested number.
        logger.error(
            "Agent %s did not respond with requested next IMA measurement list entry %s but started at %s",
            agent_id,
            agentAttestState.get_next_ima_ml_entry(),
            ima_measurement_list_entry,
        )
        failure.add_event(
            "invalid_ima_entry_nb",
            {
                "message": "Agent did not respond with requested next IMA measurement list entry",
                "got": ima_measurement_list_entry,
                "expected": agentAttestState.get_next_ima_ml_entry(),
            },
            False,
        )
    elif not agentAttestState.is_expected_boottime(boottime):
        # agent sent a list not starting at 0 and provided a boottime that doesn't
        # match the expected boottime, so it must have been rebooted; we would fail
        # attestation this time so we retry with a full attestation next time.
        agentAttestState.reset_ima_attestation()
        return failure

    agentAttestState.set_boottime(boottime)

    ima_keyrings = agentAttestState.get_ima_keyrings()
    tenant_keyring = file_signatures.ImaKeyring.from_string(agent["ima_sign_verification_keys"])
    ima_keyrings.set_tenant_keyring(tenant_keyring)

    if "tpm_clockinfo" in agent:
        agentAttestState.set_tpm_clockinfo(TPMClockInfo.from_dict(agent["tpm_clockinfo"]))

    quote_validation_failure = get_tpm_instance().check_quote(
        agentAttestState,
        agent["nonce"],
        received_public_key,
        quote,
        agent["ak_tpm"],
        agent["tpm_policy"],
        ima_measurement_list,
        ima_policy.ima_policy,
        algorithms.Hash(hash_alg),
        ima_keyrings,
        mb_measurement_list,
        agent["mb_refstate"],
        compressed=(agent["supported_version"] == "1.0"),
    )  # TODO: change this to always False after initial update
    failure.merge(quote_validation_failure)

    agent["last_received_quote"] = int(time.time())

    if not failure:
        agent["attestation_count"] += 1

        agent["tpm_clockinfo"] = agentAttestState.get_tpm_clockinfo().to_dict()

        # has public key changed? if so, clear out b64_encrypted_V, it is no longer valid
        if received_public_key != agent.get("public_key", ""):
            agent["public_key"] = received_public_key
            agent["b64_encrypted_V"] = ""
            agent["provide_V"] = True

    # ok we're done
    return failure


def prepare_v(agent):
    # be very careful printing K, U, or V as they leak in logs stored on unprotected disks
    if config.INSECURE_DEBUG:
        logger.debug("b64_V (non encrypted): %s", agent["v"])

    if agent.get("b64_encrypted_V", "") != "":
        b64_encrypted_V = agent["b64_encrypted_V"]
        logger.debug("Re-using cached encrypted V")
    else:
        # encrypt V with the public key
        b64_encrypted_V = base64.b64encode(
            crypto.rsa_encrypt(crypto.rsa_import_pubkey(agent["public_key"]), base64.b64decode(agent["v"]))
        )
        agent["b64_encrypted_V"] = b64_encrypted_V

    # logger.debug("b64_encrypted_V:" + b64_encrypted_V)
    post_data = {"encrypted_key": b64_encrypted_V}
    return post_data


def prepare_get_quote(agent):
    """This method encapsulates the action required to invoke a quote request on the Cloud Agent.

    This method is part of the polling loop of the thread launched on Tenant POST.
    """
    agentAttestState = get_AgentAttestStates().get_by_agent_id(agent["agent_id"])
    agent["nonce"] = TPM_Utilities.random_password(20)

    tpm_policy = ast.literal_eval(agent["tpm_policy"])
    params = {
        "nonce": agent["nonce"],
        "mask": tpm_policy["mask"],
        "ima_ml_entry": agentAttestState.get_next_ima_ml_entry(),
    }
    return params


def process_get_status(agent):
    agent_id = agent.agent_id
    allowlist_json = json.loads(agent.ima_policy.ima_policy)
    if isinstance(allowlist_json, dict) and "allowlist" in allowlist_json:
        al_len = len(allowlist_json["allowlist"])
    else:
        al_len = 0

    try:
        mb_refstate = json.loads(agent.mb_refstate)
    except Exception as e:
        logger.warning(
            'Non-fatal problem ocurred while attempting to evaluate agent %s attribute "mb_refstate" (%s). Will just consider the value of this attribute to be "None"',
            agent_id,
            e.args,
        )
        mb_refstate = None
        logger.debug('The contents of the agent %s attribute "mb_refstate" are %s', agent_id, agent.mb_refstate)

    if isinstance(mb_refstate, dict) and "mb_refstate" in mb_refstate:
        mb_refstate_len = len(mb_refstate["mb_refstate"])
    else:
        mb_refstate_len = 0
    response = {
        "operational_state": agent.operational_state,
        "v": agent.v,
        "ip": agent.ip,
        "port": agent.port,
        "tpm_policy": agent.tpm_policy,
        "meta_data": agent.meta_data,
        "allowlist_len": al_len,
        "mb_refstate_len": mb_refstate_len,
        "accept_tpm_hash_algs": agent.accept_tpm_hash_algs,
        "accept_tpm_encryption_algs": agent.accept_tpm_encryption_algs,
        "accept_tpm_signing_algs": agent.accept_tpm_signing_algs,
        "hash_alg": agent.hash_alg,
        "enc_alg": agent.enc_alg,
        "sign_alg": agent.sign_alg,
        "verifier_id": agent.verifier_id,
        "verifier_ip": agent.verifier_ip,
        "verifier_port": agent.verifier_port,
        "severity_level": agent.severity_level,
        "last_event_id": agent.last_event_id,
        "attestation_count": agent.attestation_count,
        "last_received_quote": agent.last_received_quote,
    }
    return response


# sign a message with revocation key.  telling of verification problem
def prepare_error(agent, msgtype="revocation", event=None):
    # prepare the revocation message:
    revocation = {
        "type": msgtype,
        "ip": agent["ip"],
        "agent_id": agent["agent_id"],
        "port": agent["port"],
        "tpm_policy": agent["tpm_policy"],
        "meta_data": agent["meta_data"],
        "event_time": time.asctime(),
    }
    if event:
        revocation["event_id"] = event.event_id
        revocation["severity_label"] = event.severity_label.name
        revocation["context"] = event.context

    tosend = {"msg": json.dumps(revocation).encode("utf-8")}

    # also need to load up private key for signing revocations
    if agent["revocation_key"] != "":
        signing_key = crypto.rsa_import_privkey(agent["revocation_key"])
        tosend["signature"] = crypto.rsa_sign(signing_key, tosend["msg"])

    else:
        tosend["signature"] = "none"
    return tosend


def validate_ima_policy_data(agent_data):
    if agent_data is None:
        return False, None

    # validate that the allowlist is proper JSON
    lists = json.loads(agent_data)

    # Validate exlude list contains valid regular expressions
    is_valid, _, err_msg = validators.valid_exclude_list(lists.get("exclude"))
    if not is_valid:
        err_msg += " Exclude list regex is misformatted. Please correct the issue and try again."

    return is_valid, err_msg



"""Add fields for revocation context to verifier

Revision ID: 257fe0f0c039
Revises: f35cdd35eb83
Create Date: 2021-08-20 12:42:30.427138

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "257fe0f0c039"
down_revision = "f35cdd35eb83"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("severity_level", sa.Integer))
    op.add_column("verifiermain", sa.Column("last_event_id", sa.String(100)))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "severity_level")
    op.drop_column("verifiermain", sa.Column("last_event_id", sa.Integer))



import unittest

from keylime import user_utils


class TestUserUtils(unittest.TestCase):
    def test_string_to_uidgid(self):
        uid, gid = user_utils.string_to_uidgid("root")
        self.assertTrue(isinstance(uid, int))
        self.assertTrue(gid is None)

        uid, gid = user_utils.string_to_uidgid("root:")
        self.assertTrue(isinstance(uid, int))
        self.assertTrue(gid is None)

        uid, gid = user_utils.string_to_uidgid(":root")
        self.assertTrue(uid is None)
        self.assertTrue(isinstance(gid, int))

        uid, gid = user_utils.string_to_uidgid("root:root")
        self.assertTrue(isinstance(uid, int))
        self.assertTrue(isinstance(gid, int))

        with self.assertRaises(ValueError):
            uid, gid = user_utils.string_to_uidgid(":")

        uid, gid = user_utils.string_to_uidgid("100")
        self.assertTrue(uid == 100)
        self.assertTrue(gid is None)

        uid, gid = user_utils.string_to_uidgid("100:")
        self.assertTrue(uid == 100)
        self.assertTrue(gid is None)

        uid, gid = user_utils.string_to_uidgid(":200")
        self.assertTrue(uid is None)
        self.assertTrue(gid == 200)

        uid, gid = user_utils.string_to_uidgid("100:200")
        self.assertTrue(uid == 100)
        self.assertTrue(gid == 200)

        with self.assertRaises(ValueError):
            uid, gid = user_utils.string_to_uidgid("-100:200")

        with self.assertRaises(ValueError):
            uid, gid = user_utils.string_to_uidgid("nobodyever")



import unittest

from keylime.tpm.tpm_main import tpm

# ############################################################
# list of input challenges for get_tpm_manufacturer function
# ############################################################
tpm_manufacturer_tests = [
    # Infineon Optiga SLB9665'
    {
        "challenge": [
            b"TPM2_PT_MANUFACTURER:\n",
            b"  aw: 0x49465800\n",
            b'  value: "IFX"\n',
            b"TPM2_PT_VENDOR_STRING_1:\n",
            b"  raw: 0x534C4239\n",
            b'  value: "SLB9"\n',
            b"TPM2_PT_VENDOR_STRING_2:\n",
            b"  raw: 0x36363500\n",
            b'  value: "665"\n',
            b"TPM2_PT_VENDOR_STRING_3:\n",
            b"  raw: 0x0\n",
            b'  value: ""\n',
            b"TPM2_PT_VENDOR_STRING_4:\n",
            b"  raw: 0x0\n",
            b'  value: ""\n',
            b"TPM2_PT_VENDOR_TPM_TYPE:\n",
            b"  raw: 0x0\n",
        ],
        "response": "SLB9",
    },
    # Nuvoton device with a wrinkle: un-escaped double quotes in the manufacturer string
    #    'Nuvoton 75x unfixed': {
    #        'challenge': [
    #            b'TPM2_PT_MANUFACTURER:\n',    b'  raw: 0x4E544300\n', b'  value: "NTC"\n',
    #            b'TPM2_PT_VENDOR_STRING_1:\n', b'  raw: 0x4E504354\n', b'  value: "NPCT"\n',
    #            b'TPM2_PT_VENDOR_STRING_3:\n', b'  raw: 0x22212134\n', b'  value: ""!!4"\n'
    #        ],
    #        'response': 'NPCT'
    #    },
    # Nuvoton 75x, assuming tpm2-tools has escaped the quote side vendor string 3'
    {
        "challenge": [
            b"TPM2_PT_MANUFACTURER:\n",
            b"  raw: 0x4E544300\n",
            b'  value: "NTC"\n',
            b"TPM2_PT_VENDOR_STRING_1:\n",
            b"  raw: 0x4E504354\n",
            b'  value: "NPCT"\n',
            b"TPM2_PT_VENDOR_STRING_3:\n",
            b"  raw: 0x22212134\n",
            b'  value: "\\"!!4"\n',
        ],
        "response": "NPCT",
    },
    # Standard software TPM
    {
        "challenge": [
            b"TPM2_PT_VENDOR_STRING_1:\n",
            b"  raw: 0x53572020\n",
            b'  value: "SW"\n',
            b"TPM2_PT_VENDOR_STRING_2:\n",
            b"  raw: 0x2054504D\n",
            b'  value: "TPM"\n',
            b"TPM2_PT_VENDOR_STRING_3:\n",
            b"  raw: 0x0\n",
            b'  value: ""\n',
            b"TPM2_PT_VENDOR_STRING_4:\n",
            b"  raw: 0x0\n",
            b'  value: ""\n',
        ],
        "response": "SW",
    },
]


class TestTPM(unittest.TestCase):
    def setUp(self):
        self.tpm = tpm()

    # basic test:
    # whatever the underlying TPM is, just ensure get_manufacturer worked.
    # we cannot predict the output on the test system, so this test merely
    # ensures that the call to get_tpm_manufacturer succeeds.
    def test_get_tpm_manufacturer(self):
        self.tpm.get_tpm_manufacturer()

    # test the challenges in the list `tpm_manufacturer_tests`
    def test_get_tpm_manufacturer_challenges(self):
        for test in tpm_manufacturer_tests:
            response = self.tpm.get_tpm_manufacturer(output=test["challenge"])
            self.assertEqual(test["response"], response)


if __name__ == "__main__":
    unittest.main()



"""AST with parser and validator for IMA ASCII entries.

Implements the templates (modes) and types as defined in:

  - https://elixir.bootlin.com/linux/latest/source/security/integrity/ima/ima_template.c
  - https://www.kernel.org/doc/html/v5.12/security/IMA-templates.html

"""

import abc
import binascii
import codecs
import struct
import typing
from typing import Callable, Dict, Optional, Union

from keylime import config, keylime_logging
from keylime.common.algorithms import Hash
from keylime.failure import Component, Failure

logger = keylime_logging.init_logging("ima")

TCG_EVENT_NAME_LEN_MAX = 255
SHA_DIGEST_LEN = 20
MD5_DIGEST_LEN = 16

NULL_BYTE = ord("\0")
COLON_BYTE = ord(":")


def get_START_HASH(hash_alg: Hash) -> bytes:
    return codecs.decode(b"0" * (hash_alg.get_size() // 4), "hex")


def get_FF_HASH(hash_alg: Hash) -> bytes:
    return codecs.decode(b"f" * (hash_alg.get_size() // 4), "hex")


class Validator:
    functions: Dict[typing.Type["Mode"], Callable[..., Failure]]

    def __init__(self, functions: Dict[typing.Type["Mode"], Callable[..., Failure]]):
        self.functions = functions

    def get_validator(self, class_type: typing.Type["Mode"]) -> Callable[..., Failure]:
        validator = self.functions.get(class_type, None)
        if validator is None:
            logger.warning("No validator was implemented for: %s. Using always false validator!", class_type)
            failure = Failure(Component.IMA, ["validation"])
            failure.add_event(
                "no_validator", f"No validator was implemented for: {class_type} . Using always false validator!", True
            )
            return lambda *_: failure
        return validator


class ParserError(TypeError):
    """Is thrown when a type could not be constructed successfully."""


class Mode(abc.ABC):
    @abc.abstractmethod
    def is_data_valid(self, validator: Validator) -> Failure:
        pass

    @abc.abstractmethod
    def bytes(self) -> bytes:
        pass


class Type(abc.ABC):
    @abc.abstractmethod
    def struct(self) -> bytes:
        pass


class HexData(Type):
    data: bytes

    def __init__(self, data: str):
        try:
            self.data = codecs.decode(data.encode("utf-8"), "hex")
        except binascii.Error as e:
            raise ParserError(f"Provided data was not valid hex: {data}") from e

    def __str__(self) -> str:
        return self.data.decode("utf-8")

    def struct(self) -> bytes:
        return struct.pack(f"<I{len(self.data)}s", len(self.data), self.data)


class Signature(HexData):
    """
    Class for type "sig".
    """

    def __init__(self, data: str):
        super().__init__(data)
        # basic checks on signature
        fmt = ">BBBIH"
        hdrlen = struct.calcsize(fmt)
        if len(self.data) < hdrlen:
            raise ParserError("Invalid signature: header too short")
        _, _, _, _, sig_size = struct.unpack(fmt, self.data[:hdrlen])

        if hdrlen + sig_size != len(self.data):
            raise ParserError("Invalid signature: malformed header")


class Buffer(HexData):
    """
    Class for type "buf".
    """


class Name(Type):
    """
    Class for type "n" and "n-ng".
    """

    name: str
    legacy: bool = False

    def __init__(self, name: str, legacy: bool = False):
        self.name = name
        self.legacy = legacy

    def __str__(self) -> str:
        return self.name

    def struct(self) -> bytes:
        name_bytes = self.name.encode("utf-8")

        # The old "n" option is fixed length.
        if self.legacy:
            return struct.pack(
                f"{len(name_bytes)}sB{TCG_EVENT_NAME_LEN_MAX - len(name_bytes)}s",
                name_bytes,
                NULL_BYTE,
                bytearray(TCG_EVENT_NAME_LEN_MAX - len(name_bytes)),
            )

        return struct.pack(f"<I{len(name_bytes)}sB", len(name_bytes) + 1, name_bytes, NULL_BYTE)


class Digest:
    """
    Class for types "d" and "d-ng" with and without algorithm
    """

    hash: bytes
    algorithm: Optional[str] = None
    legacy: bool = False

    def __init__(self, digest: str, legacy: bool = False):
        self.legacy = legacy
        tokens = digest.split(":")
        if len(tokens) == 1:
            try:
                self.hash = codecs.decode(tokens[0].encode("utf-8"), "hex")
            except binascii.Error as e:
                raise ParserError(f"Digest hash is not valid hex. Got: {tokens[0]}") from e
            if not (len(self.hash) == SHA_DIGEST_LEN or len(self.hash) == MD5_DIGEST_LEN):
                raise ParserError(
                    "Cannot create Digest. No hash algorithm is provided and hash does not belong to a md5 or sha1 hash."
                )
        elif len(tokens) == 2:
            try:
                self.hash = codecs.decode(tokens[1].encode("utf-8"), "hex")
            except binascii.Error as e:
                raise ParserError(f"Digest hash is not valid hex. Got: {tokens[1]}") from e
            self.algorithm = tokens[0]
        else:
            raise ParserError(f"Cannot create Digest expected 1 or 2 tokens got: {len(tokens)} for {digest}")

    def struct(self) -> bytes:
        # The legacy format "d" has fixed length, so it does not contain a length attribute
        if self.legacy:
            return struct.pack(f"<{len(self.hash)}s", self.hash)

        if self.algorithm is None:
            return struct.pack(f"<I{len(self.hash)}s", len(self.hash), self.hash)
        # After the ':' must be a '\O':
        # https://elixir.bootlin.com/linux/v5.12.10/source/security/integrity/ima/ima_template_lib.c#L230
        return struct.pack(
            f"<I{len(self.algorithm)}sBB{len(self.hash)}s",
            len(self.algorithm) + 2 + len(self.hash),
            self.algorithm.encode("utf-8"),
            COLON_BYTE,
            NULL_BYTE,
            self.hash,
        )

    def __eq__(self, other):
        return (
            isinstance(other, Digest)
            and self.legacy == other.legacy
            and self.hash == other.hash
            and self.algorithm == other.algorithm
        )


class Ima(Mode):
    """
    Class for "ima". Contains the digest and a path.
    """

    digest: Digest
    path: Name

    def __init__(self, data: str):
        tokens = data.split(" ", maxsplit=1)
        if len(tokens) != 2:
            raise ParserError()
        self.digest = Digest(tokens[0], legacy=True)
        self.path = Name(tokens[1], legacy=True)

    def bytes(self) -> bytes:
        return self.digest.struct() + self.path.struct()

    def is_data_valid(self, validator: Validator) -> Failure:
        return validator.get_validator(type(self))(self.digest, self.path)


class ImaNg(Mode):
    """
    Class for "ima-ng". Contains the digest and a path.
    """

    digest: Digest
    path: Name

    def __init__(self, data: str):
        tokens = data.split(" ", maxsplit=1)
        if len(tokens) != 2:
            raise ParserError(f"Cannot create ImaNg expected 2 tokens got: {len(tokens)}.")
        self.digest = Digest(tokens[0])
        self.path = Name(tokens[1])

    def bytes(self) -> bytes:
        return self.digest.struct() + self.path.struct()

    def is_data_valid(self, validator: Validator) -> Failure:
        return validator.get_validator(type(self))(self.digest, self.path)


class ImaSig(Mode):
    """
    Class for "ima-sig" template. Nearly the same as ImaNg but can contain a optional signature.
    """

    digest: Digest
    path: Name
    signature: Optional[Signature] = None

    def __init__(self, data: str):
        # There are always 3 fields in a valid entry, because:
        #
        # - n-ng does not contain a space character, as it is escaped to an
        #   underscore ("_"):
        #   https://elixir.bootlin.com/linux/v5.18/source/security/integrity/ima/ima_template_lib.c#L55
        # - the last field could be an empty string but a delimiter (" ") should
        #   still present:
        #   https://elixir.bootlin.com/linux/v5.18/source/security/integrity/ima/ima_fs.c#L244
        tokens = data.split(" ", maxsplit=2)
        if len(tokens) != 3:
            raise ParserError(f"Cannot create ImaSig expected 3 tokens got: {len(tokens)}.")
        self.digest = Digest(tokens[0])
        self.path = Name(tokens[1])
        self.signature = self.create_signature(tokens[2])

    @staticmethod
    def create_signature(hexstring: str) -> Optional[Signature]:
        """Create the Signature object if the hexstring is a valid signature"""
        try:
            return Signature(hexstring)
        except ParserError:
            pass
        return None

    def bytes(self) -> bytes:
        output = self.digest.struct() + self.path.struct()
        # If no signature is there we sill have to add the entry for it
        if self.signature is None:
            output += struct.pack("<I0s", 0, b"")
        else:
            output += self.signature.struct()
        return output

    def is_data_valid(self, validator: Validator) -> Failure:
        return validator.get_validator(type(self))(self.digest, self.path, self.signature)


class ImaBuf(Mode):
    """
    Class for "ima-buf". Contains a digest, buffer name and the buffer itself.
    For validation the buffer must be done based on the name because IMA only provides it as an byte array.
    """

    digest: Digest
    name: Name
    data: Buffer

    def __init__(self, data: str):
        # There are always 3 fields in a valid entry, because:
        #
        # - n-ng does not contain a space character, as it is escaped to an
        #   underscore ("_"):
        #   https://elixir.bootlin.com/linux/v5.18/source/security/integrity/ima/ima_template_lib.c#L55
        # - the last field could be an empty string but a delimiter (" ") should
        #   still present:
        #   https://elixir.bootlin.com/linux/v5.18/source/security/integrity/ima/ima_fs.c#L244
        tokens = data.split(" ", maxsplit=2)
        if len(tokens) != 3:
            raise ParserError(f"Cannot create ImaBuf expected 3 tokens got: {len(tokens)}.")
        self.digest = Digest(tokens[0])
        self.name = Name(tokens[1])
        self.data = Buffer(tokens[2])

    def bytes(self) -> bytes:
        return self.digest.struct() + self.name.struct() + self.data.struct()

    def is_data_valid(self, validator: Validator) -> Failure:
        return validator.get_validator(type(self))(self.digest, self.name, self.data)


class Entry:
    """
    IMA Entry. Contains the PCR, template hash and mode.
    """

    pcr: str
    ima_template_hash: bytes
    pcr_template_hash: Optional[bytes]
    mode: Mode
    _bytes: bytes
    _validator: Optional[Validator]
    _ima_hash_alg: Hash
    _pcr_hash_alg: Hash

    _mode_lookup: Dict[str, Union[typing.Type[Ima], typing.Type[ImaNg], typing.Type[ImaSig], typing.Type[ImaBuf]]] = {
        "ima": Ima,
        "ima-ng": ImaNg,
        "ima-sig": ImaSig,
        "ima-buf": ImaBuf,
    }

    def __init__(
        self,
        data: str,
        validator: Optional[Validator] = None,
        ima_hash_alg: Hash = Hash.SHA1,
        pcr_hash_alg: Hash = Hash.SHA1,
    ):
        self._validator = validator
        self._ima_hash_alg = ima_hash_alg
        self._pcr_hash_alg = pcr_hash_alg
        tokens = data.split(" ", maxsplit=3)
        if len(tokens) != 4:
            raise ParserError(f"Cannot create Entry expected 4 tokens got: {len(tokens)}.")
        self.pcr = tokens[0]
        try:
            self.ima_template_hash = codecs.decode(tokens[1].encode(), "hex")
        except binascii.Error as e:
            raise ParserError(f"Cannot create Entry expected 4 tokens got: {len(tokens)}.") from e

        mode = self._mode_lookup.get(tokens[2], None)
        if mode is None:
            raise ParserError(f"No parser for mode {tokens[2]} implemented.")
        self.mode = mode(tokens[3])
        self._bytes = self.mode.bytes()
        self.pcr_template_hash = self._pcr_hash_alg.hash(self._bytes)
        # Set correct hash for time of measure, time of use (ToMToU) errors
        # and if a file is already opened for write.
        # https://elixir.bootlin.com/linux/v5.12.12/source/security/integrity/ima/ima_main.c#L101
        if self.ima_template_hash == get_START_HASH(ima_hash_alg):
            self.ima_template_hash = get_FF_HASH(ima_hash_alg)
            self.pcr_template_hash = get_FF_HASH(pcr_hash_alg)

    def invalid(self) -> Failure:
        failure = Failure(Component.IMA, ["validation"])
        if self.pcr != str(config.IMA_PCR):
            logger.warning("IMA entry PCR does not match %s. It was: %s", config.IMA_PCR, self.pcr)
            failure.add_event(
                "ima_pcr",
                {"message": "IMA PCR is not the configured one", "expected": str(config.IMA_PCR), "got": self.pcr},
                True,
            )

        # Ignore template hash for ToMToU errors
        if self.ima_template_hash == get_FF_HASH(self._ima_hash_alg):
            logger.warning("Skipped template_hash validation entry with FF_HASH")
            # By default ToMToU errors are not treated as a failure
            if config.getboolean("verifier", "ignore_tomtou_errors", fallback=True):
                failure.add_event("tomtou", "hash validation was skipped", True)
            return failure
        if self.ima_template_hash != self._ima_hash_alg.hash(self._bytes):
            failure.add_event(
                "ima_hash",
                {
                    "message": "IMA template hash does not match the calculated hash.",
                    "expected": str(self.ima_template_hash),
                    "got": str(self.mode.bytes()),
                },
                True,
            )
            return failure
        if self._validator is None:
            failure.add_event("no_validator", "No validator specified", True)
            return failure

        failure.merge(self.mode.is_data_valid(self._validator))
        return failure



import ast


def adjust(config, mapping):
    """
    Process the configuration intermediary representation adjusting some of the
    values following changes to the configuration files semantics.

    For example, deciding values for some of the options depending on the
    content of the values in other options.  This is specially useful when some
    of the options were removed from one configuration version to the next.

    This is executed after the original configuration file is parsed, but before
    the values are replaced in the templates
    """

    print(f"Adjusting configuration")

    # Dictionary defining values to replace
    replace = {
        "components": {
            "verifier": {
                "client_cert": {"CV": "default"},
                "trusted_server_ca": {"CV": "default"},
                "database_url": {"": "sqlite"},
            },
            "registrar": {"tls_dir": {"CV": "default"}, "database_url": {"": "sqlite"}},
        }
    }

    # Dictionary defining values to convert to lists
    tolist = {
        "components": {
            "agent": [
                "trusted_client_ca",
                "revocation_actions",
            ],
            "verifier": [
                "enabled_revocation_notifications",
                "trusted_server_ca",
                "severity_labels",
                "severity_policy",
                "measured_boot_imports",
            ],
            "tenant": [
                "trusted_server_ca",
                "accept_tpm_hash_algs",
                "accept_tpm_encryption_algs",
                "accept_tpm_signing_algs",
            ],
            "registrar": ["trusted_client_ca"],
        }
    }

    for component in replace["components"].keys():
        options = replace["components"][component]
        # For each option
        for option in options.keys():
            values = options[option]
            # For each value to replace
            for to_replace in values.keys():
                # If the value in the configuration matches the one to replace
                if config[component] and config[component][option] == to_replace:
                    # Replace the value
                    config[component][option] = values[to_replace]
                    print(f'[{component}] In "{option}", replaced "{to_replace}" ' f'with "{values[to_replace]}"')

    for component in tolist["components"].keys():
        for option in tolist["components"][component]:
            # Get raw string value
            value = config[component][option]

            if value == "default":
                continue

            try:
                v = ast.literal_eval(value)
                # If the value in the config was already a list, continue
                if isinstance(v, list):
                    continue

                # If the value in the config was tuple
                if isinstance(v, tuple):
                    config[component][option] = f"{list(v)}"

            except Exception as e:
                print(
                    f"[{component}] In option '{option}', failed to parse "
                    f"'{value}' as python type, trying manual splitting"
                )

                # Eliminate leading and trailing brackets if present
                v = value.strip("[]").split(",")

                # Eliminate surrounding blank space from each element
                map(lambda x: x.strip(), v)

                # Remove empty strings
                v = list(filter(lambda x: (x != ""), v))

                config[component][option] = f"{v}"

            print(f"[{component}] For option '{option}', converted '{value}' to " f"'{config[component][option]}'")

    # Other special adjustments
    if config["verifier"]["tls_dir"] == "generate":
        for o in ["client_key", "client_cert", "trusted_server_ca", "server_key", "server_cert", "trusted_client_ca"]:
            if not config["verifier"][o] == "default":
                config["verifier"][o] = "default"
                print(f"[verifier] Replaced option '{o}' with 'default' as " f"'tls_dir' is set as 'generate'")

    if config["registrar"]["tls_dir"] == "generate":
        for o in ["server_key", "server_cert", "trusted_client_ca"]:
            if not config["registrar"][o] == "default":
                config["registrar"][o] = "default"
                print(f"[registrar] Replaced option '{o}' with 'default' as " f"'tls_dir' is set as 'generate'")

    # If the tenant's 'trusted_server_ca' is set as 'default' and both the
    # verifier's and registrar's 'tls_dir' are set as 'generate', assume the
    # user is trying to run locally and set the tenant's 'trusted_server_ca' to
    # include both CA certificate's paths.
    #
    # Note: If 'generate' is set for the registrar's 'tls_dir', the 'registrar'
    # will generate a separate CA from the verifier, meaning the user has to
    # manually set the tenant's 'trusted_server_ca' to include both locations.
    #
    # To make the registrar to use the verifier CA, set the registrar's
    # 'tls_dir' as 'default' instead of 'generate'
    if config["tenant"]["tls_dir"] == "default":
        if config["tenant"]["trusted_server_ca"] == "default":
            if config["registrar"]["tls_dir"] == "generate":
                if config["verifier"]["tls_dir"] == "generate":
                    config["tenant"][
                        "trusted_server_ca"
                    ] = "['/var/lib/keylime/cv_ca/cacert.crt', '/var/lib/keylime/reg_ca/cacert.crt']"
                    print(
                        f"[tenant] For option 'trusted_server_ca', replaced 'default' with "
                        f"'{config['tenant']['trusted_server_ca']}'"
                    )



from sqlalchemy import Column, Integer, PickleType, String, Text
from sqlalchemy.ext.declarative import declarative_base

from keylime.json import JSONPickler

Base = declarative_base()


class JSONPickleType(PickleType):  # pylint: disable=abstract-method
    impl = Text
    cache_ok = True


class RegistrarMain(Base):
    __tablename__ = "registrarmain"
    agent_id = Column(String(80), primary_key=True)
    key = Column(String(45))
    aik_tpm = Column(String(500))
    ekcert = Column(String(2048))
    ek_tpm = Column(String(500))
    mtls_cert = Column(String(2048), nullable=True)
    virtual = Column(Integer)
    ip = Column(String(15), nullable=True)
    port = Column(Integer, nullable=True)
    active = Column(Integer)
    provider_keys = Column(JSONPickleType(pickler=JSONPickler))
    regcount = Column(Integer)



"""add supported_version to verifiermain table

Revision ID: 9169f80345ed
Revises: a79c27ec1054
Create Date: 2022-01-11 20:54:00.299250

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9169f80345ed"
down_revision = "a79c27ec1054"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    with op.batch_alter_table("verifiermain") as batch_op:
        # Using server_default to assign old rows a valid value
        batch_op.add_column(sa.Column("supported_version", sa.String(length=50), nullable=False, default="1.0"))
        batch_op.alter_column("supported_version", server_default=None)


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "supported_version")



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2022 Red Hat, Inc.
"""

import unittest
from pathlib import Path

from keylime import signing

# Useful constants for the test
PACKAGE_ROOT = Path(__file__).parents[1]
TEST_FILES = f"{PACKAGE_ROOT}/test-data/files"


class TestSigning(unittest.TestCase):
    def test_sign_gpg(self):
        try:
            signing.verify_signature_from_file(
                f"{TEST_FILES}/allowlist-pgp-key.pgp",
                f"{TEST_FILES}/allowlist.json",
                f"{TEST_FILES}/allowlist-pgp-sig.sig",
                "Testing Allowlist",
            )
        except Exception as e:
            self.fail(f"Signing raised exception: {e}!")

    def test_sign_ec(self):
        try:
            signing.verify_signature_from_file(
                f"{TEST_FILES}/allowlist-ec-key.pem",
                f"{TEST_FILES}/allowlist.json",
                f"{TEST_FILES}/allowlist-ec-sig.bin",
                "Testing Allowlist",
            )
        except Exception as e:
            self.fail(f"Signing raised exception: {e}!")

    def test_sign_bad_sig(self):
        try:
            signing.verify_signature_from_file(
                f"{TEST_FILES}/allowlist-pgp-key.pgp",
                f"{TEST_FILES}/allowlist.json",
                f"{TEST_FILES}/allowlist-invalid-sig.sig",
                "Testing Allowlist",
            )
            self.fail("Signing passed with invalid signature!")
        except Exception:
            pass



"""Move (v)tpm_policy to JSONPickleType

Revision ID: f35cdd35eb83
Revises: 7d5db1a6ffb0
Create Date: 2021-08-02 15:26:34.427156

"""
import sqlalchemy as sa
from alembic import op

import keylime

# revision identifiers, used by Alembic.
revision = "f35cdd35eb83"
down_revision = "7d5db1a6ffb0"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    with op.batch_alter_table("verifiermain") as batch_op:
        batch_op.alter_column(
            "tpm_policy",
            existing_type=sa.String(1000),
            type_=keylime.db.verifier_db.JSONPickleType(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "vtpm_policy",
            existing_type=sa.String(1000),
            type_=keylime.db.verifier_db.JSONPickleType(),
            existing_nullable=True,
        )


def downgrade_cloud_verifier():
    with op.batch_alter_table("verifiermain") as batch_op:
        batch_op.alter_column(
            "tpm_policy",
            type_=sa.String(1000),
            existing_type=keylime.db.verifier_db.JSONPickleType(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "vtpm_policy",
            type_=sa.String(1000),
            existing_type=keylime.db.verifier_db.JSONPickleType(),
            existing_nullable=True,
        )



import os
import subprocess
import time

EXIT_SUCESS = 0


def _execute(cmd, env=None, **kwargs):
    with subprocess.Popen(cmd, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs) as proc:
        out, err = proc.communicate()
        return out, err, proc.returncode


def run(cmd, expectedcode=EXIT_SUCESS, raiseOnError=True, outputpaths=None, env=None, **kwargs):
    """Execute external command.

    :param cmd: a sequence of command arguments
    """
    if env is None:
        env = os.environ

    t0 = time.time()
    retout, reterr, code = _execute(cmd, env=env, **kwargs)

    t1 = time.time()
    timing = {"t1": t1, "t0": t0}

    # Gather subprocess response data
    retout_list = retout.splitlines(keepends=True)
    reterr_list = reterr.splitlines(keepends=True)

    # Don't bother continuing if call failed and we're raising on error
    if code != expectedcode and raiseOnError:
        raise Exception(
            f"Command: {cmd} returned {code}, expected {expectedcode}, " f"output {reterr_list}, stderr {reterr_list}"
        )

    # Prepare to return their file contents (if requested)
    fileouts = {}
    if isinstance(outputpaths, str):
        outputpaths = [outputpaths]
    if isinstance(outputpaths, list):
        for thispath in outputpaths:
            with open(thispath, "rb") as f:
                fileouts[thispath] = f.read()

    returnDict = {
        "retout": retout_list,
        "reterr": reterr_list,
        "code": code,
        "fileouts": fileouts,
        "timing": timing,
    }
    return returnDict


# list_contains_substring checks whether a substring is contained in the given
# list. The list may be the reterr from 'run' and contains bytes-like objects.
def list_contains_substring(lst, substring):
    for s in lst:
        if substring in str(s):
            return True
    return False



import ssl

from tornado import httpclient

from keylime import json


async def request(method, url, params=None, data=None, context=None, headers=None, timeout=60.0):

    http_client = httpclient.AsyncHTTPClient()
    if params is not None and len(list(params.keys())) > 0:
        url += "?"
        for key in list(params.keys()):
            url += f"{key}={params[key]}&"
        url = url[:-1]

    if context is not None:
        url = url.replace("http://", "https://", 1)

    # Convert dict to JSON before sending
    if isinstance(data, dict):
        data = json.dumps(data)
        if headers is None:
            headers = {}
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

    try:
        req = httpclient.HTTPRequest(
            url=url,
            method=method,
            ssl_options=context,
            body=data,
            headers=headers,
            request_timeout=timeout,
        )
        response = await http_client.fetch(req)

    except httpclient.HTTPError as e:
        if e.response is None:
            return TornadoResponse(500, str(e))

        return TornadoResponse(e.response.code, e.response.body)
    except ConnectionError as e:
        return TornadoResponse(599, f"Connection error: {str(e)}")
    except ssl.SSLError as e:
        return TornadoResponse(599, f"SSL connection error: {str(e)}")
    except OSError as e:
        return TornadoResponse(599, f"TCP/IP Connection error: {str(e)}")
    if response is None:
        return None
    return TornadoResponse(response.code, response.body)


class TornadoResponse:
    def __init__(self, code, body):
        self.status_code = code
        self.body = body



"""Validators module."""

import re


def valid_regex(regex):
    """Check if string is a valid regular expression."""
    if regex is None:
        return True, None, None

    try:
        compiled_regex = re.compile(regex)
    except re.error as regex_err:
        err = "Invalid regex: " + regex_err.msg + "."
        return False, None, err

    return True, compiled_regex, None


def valid_exclude_list(exclude_list):
    """Check if the list is composed of valid regex."""
    if not exclude_list:
        return True, None, None

    combined_regex = "(" + ")|(".join(exclude_list) + ")"
    return valid_regex(combined_regex)


def valid_hex(value):
    """Check if the string is a valid hex number representation."""
    try:
        int(value, 16)
    except Exception:
        return False
    return True


def valid_uuid(uuid: str) -> bool:
    """Check if the string is a valid UUID."""
    valid = False
    try:
        valid = bool(
            re.fullmatch(
                r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}",
                uuid,
                re.I,
            )
        )
    except Exception:
        pass
    return valid


def valid_agent_id(agent_id: str) -> bool:
    """Check if agent_id is valid."""
    valid = False
    try:
        valid = bool(re.fullmatch(r"[\w.-]+", agent_id))
    except Exception:
        pass
    return valid



import re

from packaging import version

CURRENT_VERSION = "2.1"
VERSIONS = ["1.0", "2.0", "2.1"]
LATEST_VERSIONS = {"1": "1.0", "2": "2.1"}
DEPRECATED_VERSIONS = ["1.0"]


def current_version():
    return CURRENT_VERSION


def latest_minor_version(v):
    v_obj = version.parse(str(v))
    if not isinstance(v_obj, version.Version):
        return "0"
    major_v = str(v_obj.major)
    if major_v in LATEST_VERSIONS:
        return LATEST_VERSIONS[major_v]

    return "0"


def all_versions():
    return VERSIONS.copy()


def is_supported_version(v):
    v_obj = version.parse(str(v))
    return v_obj.base_version in VERSIONS


def is_deprecated_version(v):
    v_obj = version.parse(str(v))
    return is_supported_version(v) and v_obj.base_version in DEPRECATED_VERSIONS


def normalize_version(v):
    v = str(v)
    v = v.strip("/")
    base_version = version.parse(v).base_version
    # if the base version is a single number, get the latest minor version
    if not "." in base_version:
        latest_minor = latest_minor_version(base_version)
        if latest_minor != "0":
            return latest_minor
    return base_version


def major(v):
    return version.parse(str(v)).major


def minor(v):
    return version.parse(str(v)).minor


def log_api_versions(logger):
    logger.info("Current API version %s", CURRENT_VERSION)
    versions = all_versions()
    versions.remove(CURRENT_VERSION)
    if versions:
        logger.info("Supported older API versions: " + ", ".join(versions))
    if DEPRECATED_VERSIONS:
        logger.info("Deprecated API versions (soon to be removed): " + ", ".join(DEPRECATED_VERSIONS))


def validate_version(v: str) -> bool:
    pattern = re.compile(r"\d.\d")
    return re.fullmatch(pattern, v) is not None
